# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'paths.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QMetaObject, QSize)
from PySide6.QtWidgets import (QGridLayout, QPushButton, QSizePolicy,
                               QSpacerItem)

from airunner.components.application.gui.widgets.paths.path_widget import PathWidget

class Ui_paths_form(object):
    def setupUi(self, paths_form):
        if not paths_form.objectName():
            paths_form.setObjectName(u"paths_form")
        paths_form.resize(498, 1113)
        self.gridLayout = QGridLayout(paths_form)
        self.gridLayout.setObjectName(u"gridLayout")
        self.ebook_path_widget = PathWidget(paths_form)
        self.ebook_path_widget.setObjectName(u"ebook_path_widget")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ebook_path_widget.sizePolicy().hasHeightForWidth())
        self.ebook_path_widget.setSizePolicy(sizePolicy)
        self.ebook_path_widget.setMinimumSize(QSize(0, 0))

        self.gridLayout.addWidget(self.ebook_path_widget, 3, 0, 1, 1)

        self.image_path_widget = PathWidget(paths_form)
        self.image_path_widget.setObjectName(u"image_path_widget")
        sizePolicy.setHeightForWidth(self.image_path_widget.sizePolicy().hasHeightForWidth())
        self.image_path_widget.setSizePolicy(sizePolicy)
        self.image_path_widget.setMinimumSize(QSize(0, 0))

        self.gridLayout.addWidget(self.image_path_widget, 2, 0, 1, 1)

        self.llama_index_path_widget = PathWidget(paths_form)
        self.llama_index_path_widget.setObjectName(u"llama_index_path_widget")
        sizePolicy.setHeightForWidth(self.llama_index_path_widget.sizePolicy().hasHeightForWidth())
        self.llama_index_path_widget.setSizePolicy(sizePolicy)
        self.llama_index_path_widget.setMinimumSize(QSize(0, 0))

        self.gridLayout.addWidget(self.llama_index_path_widget, 5, 0, 1, 1)

        self.base_path_widget = PathWidget(paths_form)
        self.base_path_widget.setObjectName(u"base_path_widget")
        sizePolicy.setHeightForWidth(self.base_path_widget.sizePolicy().hasHeightForWidth())
        self.base_path_widget.setSizePolicy(sizePolicy)
        self.base_path_widget.setMinimumSize(QSize(0, 0))

        self.gridLayout.addWidget(self.base_path_widget, 1, 0, 1, 1)

        self.documents_path_widget = PathWidget(paths_form)
        self.documents_path_widget.setObjectName(u"documents_path_widget")
        sizePolicy.setHeightForWidth(self.documents_path_widget.sizePolicy().hasHeightForWidth())
        self.documents_path_widget.setSizePolicy(sizePolicy)
        self.documents_path_widget.setMinimumSize(QSize(0, 0))

        self.gridLayout.addWidget(self.documents_path_widget, 4, 0, 1, 1)

        self.auto_button = QPushButton(paths_form)
        self.auto_button.setObjectName(u"auto_button")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.auto_button.sizePolicy().hasHeightForWidth())
        self.auto_button.setSizePolicy(sizePolicy1)
        self.auto_button.setMinimumSize(QSize(0, 0))

        self.gridLayout.addWidget(self.auto_button, 0, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 6, 0, 1, 1)


        self.retranslateUi(paths_form)
        self.auto_button.clicked.connect(paths_form.action_button_clicked_reset)

        QMetaObject.connectSlotsByName(paths_form)
    # setupUi

    def retranslateUi(self, paths_form):
        paths_form.setWindowTitle(QCoreApplication.translate("paths_form", u"Form", None))
        self.ebook_path_widget.setProperty(u"title", QCoreApplication.translate("paths_form", u"Ebook Path", None))
        self.ebook_path_widget.setProperty(u"description", QCoreApplication.translate("paths_form", u"Absolute path to directory containing ebooks (epub)", None))
        self.ebook_path_widget.setProperty(u"path_name", QCoreApplication.translate("paths_form", u"ebook_path", None))
        self.image_path_widget.setProperty(u"title", QCoreApplication.translate("paths_form", u"Image Path", None))
        self.image_path_widget.setProperty(u"description", QCoreApplication.translate("paths_form", u"Absolute path to the directory which will contain generated images", None))
        self.image_path_widget.setProperty(u"path_name", QCoreApplication.translate("paths_form", u"image_path", None))
        self.llama_index_path_widget.setProperty(u"title", QCoreApplication.translate("paths_form", u"LLama Index Path", None))
        self.llama_index_path_widget.setProperty(u"description", QCoreApplication.translate("paths_form", u"Absolute path to directory which will store files for LLama Index", None))
        self.llama_index_path_widget.setProperty(u"path_name", QCoreApplication.translate("paths_form", u"llama_index_path", None))
        self.base_path_widget.setProperty(u"title", QCoreApplication.translate("paths_form", u"Base Path", None))
        self.base_path_widget.setProperty(u"description", QCoreApplication.translate("paths_form", u"Absolute path to the directory which will hold all model directories", None))
        self.base_path_widget.setProperty(u"path_name", QCoreApplication.translate("paths_form", u"base_path", None))
        self.documents_path_widget.setProperty(u"title", QCoreApplication.translate("paths_form", u"Documents Path", None))
        self.documents_path_widget.setProperty(u"description", QCoreApplication.translate("paths_form", u"Absolute path to directory containing documnets (txt, PDF)", None))
        self.documents_path_widget.setProperty(u"path_name", QCoreApplication.translate("paths_form", u"documents_path", None))
        self.auto_button.setText(QCoreApplication.translate("paths_form", u"Reset Paths to Default Values", None))
    # retranslateUi




"""Integration tests for filter system with database."""

import pytest
from PIL import Image
import numpy as np

from airunner.components.art.data.image_filter import ImageFilter
from airunner.components.art.data.image_filter_value import ImageFilterValue
from airunner.components.art.utils.image_filter_utils import (
    get_filter_values,
    build_filter_instance,
)


class TestFilterDatabaseIntegration:
    """Integration tests for filter system with database."""

    @pytest.fixture
    def test_image(self):
        """Create a test image."""
        img_array = np.random.randint(0, 255, (100, 100, 3), dtype=np.uint8)
        return Image.fromarray(img_array, "RGB")

    def test_pixel_art_filter_from_database(self, test_image):
        """Test loading and applying pixel_art filter from database."""
        # Get filter from database
        filters = ImageFilter.objects.filter_by(name="pixel_art")

        if not filters:
            pytest.skip("pixel_art filter not in database")

        filter_obj = filters[0]

        # Get filter values
        filter_values = get_filter_values(filter_obj.id)

        assert len(filter_values) > 0, "No filter values found"

        # Build filter instance
        filter_instance = build_filter_instance("pixel_art")

        assert filter_instance is not None, "Failed to build filter instance"

        # Apply filter
        result = filter_instance.filter(test_image)

        assert result is not None
        assert isinstance(result, Image.Image)
        assert result.size == test_image.size

    def test_all_filters_can_be_built(self):
        """Test that all filters in database can be instantiated."""
        all_filters = ImageFilter.objects.all()

        failed_filters = []

        for filter_obj in all_filters:
            try:
                instance = build_filter_instance(filter_obj.name)
                if instance is None:
                    failed_filters.append(
                        (
                            filter_obj.name,
                            "build_filter_instance returned None",
                        )
                    )
            except Exception as e:
                failed_filters.append((filter_obj.name, str(e)))

        assert (
            len(failed_filters) == 0
        ), f"Failed to build filters: {failed_filters}"

    def test_pixel_art_filter_parameters_are_valid(self):
        """Test pixel_art filter parameters have correct types and ranges."""
        filters = ImageFilter.objects.filter_by(name="pixel_art")

        if not filters:
            pytest.skip("pixel_art filter not in database")

        filter_obj = filters[0]
        values = ImageFilterValue.objects.filter_by(
            image_filter_id=filter_obj.id
        )

        # Expected parameters
        expected_params = {
            "number_of_colors": {"type": "int", "min": 2, "max": 1024},
            "base_size": {"type": "int", "min": 2, "max": 256},
            "smoothing": {"type": "int", "min": 0, "max": 100},
        }

        for value in values:
            if value.name in expected_params:
                expected = expected_params[value.name]

                # Check type
                assert (
                    value.value_type == expected["type"]
                ), f"{value.name} should be {expected['type']}, got {value.value_type}"

                # Check min/max are set
                assert (
                    value.min_value is not None
                ), f"{value.name} missing min_value"
                assert (
                    value.max_value is not None
                ), f"{value.name} missing max_value"

                # Check min/max values are correct
                min_val = float(value.min_value)
                max_val = float(value.max_value)

                assert (
                    min_val == expected["min"]
                ), f"{value.name} min should be {expected['min']}, got {min_val}"
                assert (
                    max_val == expected["max"]
                ), f"{value.name} max should be {expected['max']}, got {max_val}"

    def test_filter_value_conversion_works(self):
        """Test that filter values convert correctly from strings."""
        filters = ImageFilter.objects.filter_by(name="pixel_art")

        if not filters:
            pytest.skip("pixel_art filter not in database")

        filter_obj = filters[0]
        values = ImageFilterValue.objects.filter_by(
            image_filter_id=filter_obj.id
        )

        for value in values:
            # All values should be stored as strings
            assert isinstance(
                value.value, str
            ), f"{value.name} value should be string, got {type(value.value)}"

            # But should be convertible to their type
            if value.value_type == "int":
                try:
                    int(value.value)
                except ValueError:
                    pytest.fail(
                        f"{value.name} value '{value.value}' not convertible to int"
                    )
            elif value.value_type == "float":
                try:
                    float(value.value)
                except ValueError:
                    pytest.fail(
                        f"{value.name} value '{value.value}' not convertible to float"
                    )

    def test_filter_with_corrupted_zero_value_fails_gracefully(
        self, test_image
    ):
        """Test that filter handles corrupted zero value gracefully."""
        # Simulate corrupted value (like the bug we fixed)
        from airunner.components.art.filters.pixel_art import PixelFilter

        # Create filter with zero colors (corrupted state)
        filter_instance = PixelFilter(
            number_of_colors=0, base_size=256, smoothing=0
        )

        # Should not crash
        result = filter_instance.filter(test_image)

        # Should return an image (even if unchanged)
        assert result is not None
        assert isinstance(result, Image.Image)



"""Tests for EventHandlerMixin."""

from unittest.mock import MagicMock, Mock, patch
from PySide6.QtCore import QEvent, QPointF, QSize, Qt, QTimer
from PySide6.QtGui import QMouseEvent, QResizeEvent, QKeyEvent
from PySide6.QtWidgets import QGraphicsScene

from airunner.components.art.gui.widgets.canvas.mixins.event_handler_mixin import (
    EventHandlerMixin,
)


class BaseStub:
    """Stub base class providing event methods that mixins call via super()."""

    def wheelEvent(self, event):
        """Stub wheelEvent."""

    def mousePressEvent(self, event):
        """Stub mousePressEvent."""

    def mouseReleaseEvent(self, event):
        """Stub mouseReleaseEvent."""

    def mouseMoveEvent(self, event):
        """Stub mouseMoveEvent."""

    def keyPressEvent(self, event):
        """Stub keyPressEvent."""

    def showEvent(self, event):
        """Stub showEvent."""

    def resizeEvent(self, event):
        """Stub resizeEvent."""

    def enterEvent(self, event):
        """Stub enterEvent."""

    def leaveEvent(self, event):
        """Stub leaveEvent."""


class TestableEventHandlerMixin(EventHandlerMixin, BaseStub):
    """Testable version of EventHandlerMixin with required dependencies."""

    def __init__(self):
        """Initialize testable mixin with mocked dependencies."""
        # Scene mock
        self._scene = MagicMock(spec=QGraphicsScene)

        # State flags
        self._middle_mouse_pressed = False
        self.last_pos = None
        self._initialized = False
        self._is_restoring_state = False
        self._pending_pan_event = False

        # Canvas offset
        self._canvas_offset = QPointF(0, 0)

        # Pan timer
        self._pan_update_timer = Mock(spec=QTimer)
        self._pan_update_timer.isActive = Mock(return_value=False)
        self._pan_update_timer.start = Mock()

        # Viewport tracking
        viewport_mock = Mock()
        viewport_mock.size.return_value = QSize(800, 600)
        self._viewport = viewport_mock
        self._last_viewport_size = QSize(800, 600)

        # Grid compensation
        self._grid_compensation_offset = QPointF(0, 0)

        # Text items
        self._text_items = []
        self._text_item_layer_map = {}

        # Logger
        self.logger = Mock()
        self.logger.info = Mock()
        self.logger.debug = Mock()
        self.logger.exception = Mock()

        # Settings
        self.settings = Mock()
        self.settings.value = Mock(side_effect=lambda key, default: default)

        # API
        self.api = Mock()
        self.api.art.canvas.update_grid_info = Mock()

        # Methods to mock
        self.draw_grid = Mock()
        self.save_canvas_offset = Mock()
        self.load_canvas_offset = Mock()
        self.do_draw = Mock()
        self.toggle_drag_mode = Mock()
        self.set_canvas_color = Mock()
        self.show_active_grid_area = Mock()
        self.align_canvas_items_to_viewport = Mock()
        self.update_active_grid_area_position = Mock()
        self.updateImagePositions = Mock()
        self._apply_viewport_compensation = Mock()
        self._remove_text_item = Mock()

        # Mock super() methods that EventHandlerMixin calls
        self._super_wheelEvent = Mock()
        self._super_mousePressEvent = Mock()
        self._super_mouseReleaseEvent = Mock()
        self._super_mouseMoveEvent = Mock()
        self._super_keyPressEvent = Mock()
        self._super_showEvent = Mock()
        self._super_resizeEvent = Mock()
        self._super_enterEvent = Mock()
        self._super_leaveEvent = Mock()

    @property
    def scene(self):
        """Scene property."""
        return self._scene

    @property
    def canvas_offset(self):
        """Canvas offset property."""
        return self._canvas_offset

    @canvas_offset.setter
    def canvas_offset(self, value):
        """Set canvas offset."""
        self._canvas_offset = value

    @property
    def canvas_offset_x(self):
        """Canvas offset X coordinate."""
        return self._canvas_offset.x()

    @property
    def canvas_offset_y(self):
        """Canvas offset Y coordinate."""
        return self._canvas_offset.y()

    def viewport(self):
        """Return viewport mock."""
        return self._viewport


class TestEventHandlerMixin:
    """Tests for EventHandlerMixin class."""

    def test_wheelEvent_with_ctrl_modifier(self, qapp):
        """Test wheelEvent allows zoom with Ctrl modifier."""
        mixin = TestableEventHandlerMixin()
        event = Mock(spec=QMouseEvent)
        event.modifiers.return_value = Qt.KeyboardModifier.ControlModifier
        event.ignore = Mock()

        # Don't patch super() - we want our code to execute
        mixin.wheelEvent(event)

        mixin.draw_grid.assert_called_once()
        event.ignore.assert_not_called()

    def test_wheelEvent_without_ctrl_modifier(self, qapp):
        """Test wheelEvent ignores scroll without Ctrl modifier."""
        mixin = TestableEventHandlerMixin()
        event = Mock(spec=QMouseEvent)
        event.modifiers.return_value = Qt.KeyboardModifier.NoModifier
        event.ignore = Mock()

        mixin.wheelEvent(event)

        event.ignore.assert_called_once()
        mixin.draw_grid.assert_not_called()

    def test_mousePressEvent_middle_button(self, qapp):
        """Test mousePressEvent handles middle mouse button press."""
        mixin = TestableEventHandlerMixin()
        event = Mock(spec=QMouseEvent)
        event.button.return_value = Qt.MouseButton.MiddleButton
        event.pos.return_value = QPointF(100, 150)
        event.accept = Mock()

        mixin.mousePressEvent(event)

        assert mixin._middle_mouse_pressed is True
        assert mixin.last_pos == QPointF(100, 150)
        event.accept.assert_called_once()

    def test_mousePressEvent_other_button(self, qapp):
        """Test mousePressEvent delegates non-middle button presses."""
        mixin = TestableEventHandlerMixin()
        event = Mock(spec=QMouseEvent)
        event.button.return_value = Qt.MouseButton.LeftButton

        with patch.object(EventHandlerMixin, "mousePressEvent"):
            mixin.mousePressEvent(event)

        assert mixin._middle_mouse_pressed is False

    def test_mouseReleaseEvent_middle_button(self, qapp):
        """Test mouseReleaseEvent handles middle button release."""
        mixin = TestableEventHandlerMixin()
        mixin._middle_mouse_pressed = True
        mixin.last_pos = QPointF(100, 100)
        mixin.scene.handle_cursor = Mock()  # Add handle_cursor to scene mock

        event = Mock(spec=QMouseEvent)
        event.button.return_value = Qt.MouseButton.MiddleButton
        event.accept = Mock()

        mixin.mouseReleaseEvent(event)

        mixin.save_canvas_offset.assert_called_once()
        assert mixin._middle_mouse_pressed is False
        assert mixin.last_pos is None
        event.accept.assert_called_once()

    def test_mouseReleaseEvent_triggers_cursor_update(self, qapp):
        """Test mouseReleaseEvent triggers scene cursor update."""
        mixin = TestableEventHandlerMixin()
        mixin._middle_mouse_pressed = True
        mixin.scene.handle_cursor = Mock()

        event = Mock(spec=QMouseEvent)
        event.button.return_value = Qt.MouseButton.MiddleButton
        event.accept = Mock()

        mixin.mouseReleaseEvent(event)

        mixin.scene.handle_cursor.assert_called_once()

    def test_mouseReleaseEvent_other_button(self, qapp):
        """Test mouseReleaseEvent delegates other button releases."""
        mixin = TestableEventHandlerMixin()
        event = Mock(spec=QMouseEvent)
        event.button.return_value = Qt.MouseButton.LeftButton

        with patch.object(EventHandlerMixin, "mouseReleaseEvent"):
            mixin.mouseReleaseEvent(event)

        mixin.save_canvas_offset.assert_not_called()

    def test_mouseMoveEvent_during_pan(self, qapp):
        """Test mouseMoveEvent handles panning with middle mouse."""
        mixin = TestableEventHandlerMixin()
        mixin._middle_mouse_pressed = True
        mixin.last_pos = QPointF(100, 100)
        mixin.canvas_offset = QPointF(50, 50)

        event = Mock()
        event.pos.return_value = QPointF(110, 105)
        event.accept = Mock()

        mixin.mouseMoveEvent(event)

        # Delta is (110-100, 105-100) = (10, 5)
        # New offset should be (50-10, 50-5) = (40, 45)
        assert mixin.canvas_offset == QPointF(40, 45)
        assert mixin.last_pos == QPointF(110, 105)
        mixin.api.art.canvas.update_grid_info.assert_called_once()
        mixin._pan_update_timer.start.assert_called_once_with(1)
        event.accept.assert_called_once()

    def test_mouseMoveEvent_during_active_pan_timer(self, qapp):
        """Test mouseMoveEvent sets pending flag when pan timer active."""
        mixin = TestableEventHandlerMixin()
        mixin._middle_mouse_pressed = True
        mixin.last_pos = QPointF(100, 100)
        mixin._pan_update_timer.isActive = Mock(return_value=True)

        event = Mock()
        event.pos.return_value = QPointF(105, 105)
        event.accept = Mock()

        mixin.mouseMoveEvent(event)

        assert mixin._pending_pan_event is True
        # Should not start timer again
        mixin._pan_update_timer.start.assert_not_called()

    def test_mouseMoveEvent_without_pan(self, qapp):
        """Test mouseMoveEvent delegates when not panning."""
        mixin = TestableEventHandlerMixin()
        mixin._middle_mouse_pressed = False
        event = Mock()

        with patch.object(EventHandlerMixin, "mouseMoveEvent"):
            mixin.mouseMoveEvent(event)

        mixin.api.art.canvas.update_grid_info.assert_not_called()

    def test_keyPressEvent_delete_key_with_selected_text(self, qapp):
        """Test keyPressEvent deletes selected text items."""
        mixin = TestableEventHandlerMixin()

        # Create mock text items
        selected_item = Mock()
        selected_item.isSelected.return_value = True
        unselected_item = Mock()
        unselected_item.isSelected.return_value = False

        mixin._text_items = [selected_item, unselected_item]
        mixin._text_item_layer_map = {selected_item: 1, unselected_item: 2}
        mixin.scene._history_transactions = {}
        mixin.scene._begin_layer_history_transaction = Mock()
        mixin.scene._commit_layer_history_transaction = Mock()

        event = Mock(spec=QKeyEvent)
        event.key.return_value = Qt.Key.Key_Delete

        mixin.keyPressEvent(event)

        # Should remove only selected item
        mixin._remove_text_item.assert_called_once_with(
            selected_item, manage_transaction=False
        )
        mixin.scene._begin_layer_history_transaction.assert_called_once_with(
            1, "text"
        )
        mixin.scene._commit_layer_history_transaction.assert_called_once_with(
            1, "text"
        )

    def test_keyPressEvent_delete_key_no_selection(self, qapp):
        """Test keyPressEvent with Delete but no selected items."""
        mixin = TestableEventHandlerMixin()
        mixin._text_items = []

        event = Mock(spec=QKeyEvent)
        event.key.return_value = Qt.Key.Key_Delete

        with patch.object(EventHandlerMixin, "keyPressEvent"):
            mixin.keyPressEvent(event)

        mixin._remove_text_item.assert_not_called()

    def test_keyPressEvent_other_key(self, qapp):
        """Test keyPressEvent delegates non-Delete keys."""
        mixin = TestableEventHandlerMixin()
        event = Mock(spec=QKeyEvent)
        event.key.return_value = Qt.Key.Key_A

        with patch.object(EventHandlerMixin, "keyPressEvent"):
            mixin.keyPressEvent(event)

        mixin._remove_text_item.assert_not_called()

    def test_do_pan_update(self, qapp):
        """Test _do_pan_update refreshes positions and grid."""
        mixin = TestableEventHandlerMixin()

        mixin._do_pan_update()

        mixin.update_active_grid_area_position.assert_called_once()
        mixin.updateImagePositions.assert_called_once()
        mixin.draw_grid.assert_called_once()

    def test_do_pan_update_with_pending_event(self, qapp):
        """Test _do_pan_update restarts timer when pending event exists."""
        mixin = TestableEventHandlerMixin()
        mixin._pending_pan_event = True

        mixin._do_pan_update()

        assert mixin._pending_pan_event is False
        mixin._pan_update_timer.start.assert_called_once_with(1)

    def test_showEvent_first_show(self, qapp):
        """Test showEvent initializes canvas on first show."""
        mixin = TestableEventHandlerMixin()
        mixin._initialized = False
        mixin.scene.original_item_positions = {}
        mixin.scene._refresh_layer_display = Mock()
        mixin.scene.show_event = Mock()

        # Mock QTimer.singleShot to execute callback immediately
        with patch(
            "airunner.components.art.gui.widgets.canvas.mixins.event_handler_mixin.QTimer.singleShot",
            side_effect=lambda delay, func: func(),
        ):
            event = Mock(spec=QEvent)
            # Don't patch super() - let showEvent execute
            mixin.showEvent(event)

        assert mixin._initialized is True
        mixin.load_canvas_offset.assert_called_once()
        mixin.do_draw.assert_called_once_with(True)
        mixin.toggle_drag_mode.assert_called_once()
        mixin.set_canvas_color.assert_called_once_with(mixin.scene)
        mixin.show_active_grid_area.assert_called_once()
        mixin.scene._refresh_layer_display.assert_called_once()
        mixin.align_canvas_items_to_viewport.assert_called_once()

    def test_showEvent_subsequent_show_no_resize(self, qapp):
        """Test showEvent on subsequent show with unchanged viewport."""
        mixin = TestableEventHandlerMixin()
        mixin._initialized = True
        mixin._last_viewport_size = QSize(800, 600)
        mixin._viewport.size.return_value = QSize(800, 600)

        event = Mock(spec=QEvent)
        with patch.object(EventHandlerMixin, "showEvent"):
            mixin.showEvent(event)

        mixin._apply_viewport_compensation.assert_not_called()
        mixin.draw_grid.assert_not_called()

    def test_showEvent_subsequent_show_with_resize(self, qapp):
        """Test showEvent compensates for viewport size change while hidden."""
        mixin = TestableEventHandlerMixin()
        mixin._initialized = True
        mixin._is_restoring_state = False
        mixin._last_viewport_size = QSize(800, 600)
        mixin._viewport.size.return_value = QSize(1000, 700)

        event = Mock(spec=QEvent)
        # Don't patch super() - let showEvent execute
        mixin.showEvent(event)

        # New center: (500, 350), old center: (400, 300)
        # Shift: (100, 50)
        mixin._apply_viewport_compensation.assert_called_once_with(100.0, 50.0)
        mixin.draw_grid.assert_called_once()
        assert mixin._last_viewport_size == QSize(1000, 700)

    def test_finish_state_restoration(self, qapp):
        """Test _finish_state_restoration completes initialization."""
        mixin = TestableEventHandlerMixin()
        mixin._is_restoring_state = True
        mixin.settings.value = Mock(
            side_effect=lambda key, default: 10.0 if "x" in key else 20.0
        )
        mixin.scene.show_event = Mock()

        mixin._finish_state_restoration()

        assert mixin._is_restoring_state is False
        assert mixin.canvas_offset == QPointF(10.0, 20.0)
        mixin.update_active_grid_area_position.assert_called_once()
        mixin.updateImagePositions.assert_called_once()
        mixin.scene.show_event.assert_called_once()

    def test_resizeEvent_during_restoration(self, qapp):
        """Test resizeEvent skips compensation during state restoration."""
        mixin = TestableEventHandlerMixin()
        mixin._is_restoring_state = True
        mixin._initialized = False

        event = Mock(spec=QResizeEvent)
        with patch.object(EventHandlerMixin, "resizeEvent"):
            mixin.resizeEvent(event)

        mixin._apply_viewport_compensation.assert_not_called()
        mixin.draw_grid.assert_not_called()

    def test_resizeEvent_after_initialization(self, qapp):
        """Test resizeEvent applies compensation after initialization."""
        mixin = TestableEventHandlerMixin()
        mixin._is_restoring_state = False
        mixin._initialized = True
        mixin._last_viewport_size = QSize(800, 600)
        mixin._viewport.size.return_value = QSize(1000, 700)

        event = Mock(spec=QResizeEvent)
        # Don't patch super() - let resizeEvent execute
        mixin.resizeEvent(event)

        # New center: (500, 350), old center: (400, 300)
        # Shift: (100, 50)
        mixin._apply_viewport_compensation.assert_called_once_with(100.0, 50.0)
        mixin.draw_grid.assert_called_once()
        assert mixin._last_viewport_size == QSize(1000, 700)

    def test_resizeEvent_no_size_change(self, qapp):
        """Test resizeEvent skips compensation when size unchanged."""
        mixin = TestableEventHandlerMixin()
        mixin._is_restoring_state = False
        mixin._initialized = True
        mixin._last_viewport_size = QSize(800, 600)
        mixin._viewport.size.return_value = QSize(800, 600)

        event = Mock(spec=QResizeEvent)
        with patch.object(EventHandlerMixin, "resizeEvent"):
            mixin.resizeEvent(event)

        mixin._apply_viewport_compensation.assert_not_called()
        mixin.draw_grid.assert_not_called()

    def test_enterEvent(self, qapp):
        """Test enterEvent delegates to scene."""
        mixin = TestableEventHandlerMixin()
        mixin.scene.enterEvent = Mock()

        event = Mock(spec=QEvent)
        # Don't patch super() - let enterEvent execute
        mixin.enterEvent(event)

        mixin.scene.enterEvent.assert_called_once_with(event)

    def test_leaveEvent(self, qapp):
        """Test leaveEvent delegates to scene."""
        mixin = TestableEventHandlerMixin()
        mixin.scene.leaveEvent = Mock()

        event = Mock(spec=QEvent)
        # Don't patch super() - let leaveEvent execute
        mixin.leaveEvent(event)

        mixin.scene.leaveEvent.assert_called_once_with(event)



"""Add xtts_settings table

Revision ID: dffe1636eeaf
Revises: 843e4b044d4d
Create Date: 2025-11-30 18:34:11.469944

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

from airunner.utils.db import drop_table


# revision identifiers, used by Alembic.
revision: str = 'dffe1636eeaf'
down_revision: Union[str, None] = '843e4b044d4d'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # Note: XTTS has been removed from the project.
    # This migration originally added an xtts_settings table, but since XTTS
    # is no longer supported, we skip creating that table.
    # We still drop the deprecated speech_t5_settings table.
    drop_table(table_name="speech_t5_settings")


def downgrade() -> None:
    # Recreate speech_t5_settings table
    op.create_table('speech_t5_settings',
    sa.Column('id', sa.INTEGER(), nullable=False),
    sa.Column('datasets_path', sa.VARCHAR(), nullable=True),
    sa.Column('processor_path', sa.VARCHAR(), nullable=True),
    sa.Column('vocoder_path', sa.VARCHAR(), nullable=True),
    sa.Column('model_path', sa.VARCHAR(), nullable=True),
    sa.Column('pitch', sa.INTEGER(), nullable=True),
    sa.Column('voice', sa.VARCHAR(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    
    # Drop xtts_settings table if it exists
    drop_table(table_name="xtts_settings")


"""
Mixin providing memory optimization for Stable Diffusion models.

This mixin handles all memory efficiency settings including VAE slicing,
attention slicing, CPU offload, ToMe SD, and other VRAM optimizations.
"""

import torch
import tomesd
from diffusers.models.attention_processor import (
    AttnProcessor,
    AttnProcessor2_0,
)

from airunner.settings import (
    AIRUNNER_MEM_USE_LAST_CHANNELS,
    AIRUNNER_MEM_USE_ENABLE_VAE_SLICING,
    AIRUNNER_MEM_USE_ATTENTION_SLICING,
    AIRUNNER_MEM_USE_TILED_VAE,
    AIRUNNER_MEM_USE_ACCELERATED_TRANSFORMERS,
    AIRUNNER_MEM_USE_ENABLE_SEQUENTIAL_CPU_OFFLOAD,
    AIRUNNER_MEM_ENABLE_MODEL_CPU_OFFLOAD,
    AIRUNNER_MEM_USE_TOME_SD,
    AIRUNNER_MEM_TOME_SD_RATIO,
    AIRUNNER_DISABLE_FLASH_ATTENTION,
)
from airunner.utils.memory import is_ampere_or_newer
from airunner.utils.settings import get_qsettings


class SDMemoryManagementMixin:
    """Mixin providing memory optimization for Stable Diffusion models."""

    def _make_memory_efficient(self):
        """
        Apply all memory efficiency settings to pipeline.

        Applies settings like VAE slicing, attention slicing, CPU offload,
        ToMe SD, xformers, and torch.compile based on current memory settings.
        """
        self._current_memory_settings = self.memory_settings.to_dict()
        if not self._pipe:
            self.logger.error("Pipe is None, unable to apply memory settings")
            return

        # Apply all memory efficient settings using the helper method
        self._apply_memory_setting(
            "last_channels_applied",
            "use_last_channels",
            self._apply_last_channels,
        )
        self._apply_memory_setting(
            "vae_slicing_applied",
            "use_enable_vae_slicing",
            self._apply_vae_slicing,
        )
        self._apply_memory_setting(
            "attention_slicing_applied",
            "use_attention_slicing",
            self._apply_attention_slicing,
        )
        self._apply_memory_setting(
            "tiled_vae_applied", "use_tiled_vae", self._apply_tiled_vae
        )
        self._apply_memory_setting(
            "accelerated_transformers_applied",
            "use_accelerated_transformers",
            self._apply_accelerated_transformers,
        )

        # Apply xformers for Flash Attention 2 (major speedup)
        self._apply_xformers()

        # Apply torch.compile for 2-3x speedup (one-time compilation)
        self._apply_torch_compile()

        self._apply_memory_setting(
            "cpu_offload_applied",
            "use_enable_sequential_cpu_offload",
            self._apply_cpu_offload,
        )
        self._apply_memory_setting(
            "model_cpu_offload_applied",
            "enable_model_cpu_offload",
            self._apply_model_offload,
        )
        self._apply_memory_setting(
            "tome_sd_applied", "use_tome_sd", self._apply_tome
        )

    def _apply_memory_setting(self, setting_name, attribute_name, apply_func):
        """
        Apply individual memory setting if changed.

        Args:
            setting_name: Name of flag tracking application state.
            attribute_name: Name of setting attribute.
            apply_func: Function to call to apply the setting.
        """
        attr_val = getattr(self.memory_settings, attribute_name)
        if self._memory_settings_flags[setting_name] != attr_val:
            apply_func(attr_val)
            self._memory_settings_flags[setting_name] = attr_val

    def _apply_last_channels(self, attr_val):
        """
        Apply channels_last memory format to UNet.

        Args:
            attr_val: Whether to enable channels_last format.

        Channels-last improves memory access patterns on some GPUs.
        """
        enabled = AIRUNNER_MEM_USE_LAST_CHANNELS
        if enabled is None:
            enabled = attr_val
        self.logger.debug(
            f"{'Enabling' if enabled else 'Disabling'} torch.channels_last"
        )
        self._pipe.unet.to(
            memory_format=(
                torch.channels_last if enabled else torch.contiguous_format
            )
        )

    def _apply_vae_slicing(self, attr_val):
        """
        Apply VAE slicing for reduced VRAM usage.

        Args:
            attr_val: Whether to enable VAE slicing.

        VAE slicing processes images in smaller tiles to reduce memory.
        """
        enabled = AIRUNNER_MEM_USE_ENABLE_VAE_SLICING
        if enabled is None:
            enabled = attr_val
        try:
            if enabled:
                self.logger.debug("Enabling vae slicing")
                self._pipe.enable_vae_slicing()
            else:
                self.logger.debug("Disabling vae slicing")
                self._pipe.disable_vae_slicing()
        except AttributeError as e:
            self.logger.error("Failed to apply vae slicing")
            self.logger.error(e)

    def _apply_attention_slicing(self, attr_val):
        """
        Apply attention slicing for reduced VRAM usage.

        Args:
            attr_val: Whether to enable attention slicing.

        Attention slicing processes attention in smaller chunks.
        """
        enabled = AIRUNNER_MEM_USE_ATTENTION_SLICING
        if enabled is None:
            enabled = attr_val
        try:
            if attr_val:
                self.logger.debug("Enabling attention slicing")
                self._pipe.enable_attention_slicing(1)
            else:
                self.logger.debug("Disabling attention slicing")
                self._pipe.disable_attention_slicing()
        except AttributeError as e:
            self.logger.warning(f"Failed to apply attention slicing: {e}")

    def _apply_tiled_vae(self, attr_val):
        """
        Apply tiled VAE for very large images.

        Args:
            attr_val: Whether to enable tiled VAE.

        Tiled VAE processes large images in overlapping tiles.
        """
        enabled = AIRUNNER_MEM_USE_TILED_VAE
        if enabled is None:
            enabled = attr_val
        try:
            if enabled:
                self.logger.debug("Enabling tiled vae")
                self._pipe.vae.enable_tiling()
            else:
                self.logger.debug("Disabling tiled vae")
                self._pipe.vae.disable_tiling()
        except AttributeError:
            self.logger.warning("Tiled vae not supported for this model")

    def _apply_accelerated_transformers(self, attr_val):
        """
        Apply accelerated transformers (attention processor 2.0).

        Args:
            attr_val: Whether to enable accelerated transformers.

        Uses optimized attention on Ampere+ GPUs (RTX 30xx/40xx).
        """
        enabled = AIRUNNER_MEM_USE_ACCELERATED_TRANSFORMERS
        if enabled is None:
            enabled = attr_val

        if not is_ampere_or_newer(self._device_index):
            enabled = False

        self.logger.debug(
            f"{'Enabling' if enabled else 'Disabling'} accelerated transformers"
        )
        self._pipe.unet.set_attn_processor(
            AttnProcessor2_0() if enabled else AttnProcessor()
        )

    def _apply_cpu_offload(self, attr_val):
        """
        Apply sequential CPU offload for maximum VRAM savings.

        Args:
            attr_val: Whether to enable sequential CPU offload.

        Moves model components to CPU when not in use. Slow but minimal VRAM.
        Disabled for SDXL + Compel due to compatibility issues.
        """
        enabled = AIRUNNER_MEM_USE_ENABLE_SEQUENTIAL_CPU_OFFLOAD
        if enabled is None:
            enabled = attr_val

        # Disable sequential CPU offload only for SDXL models when compel is enabled
        # due to compatibility issues with compel prompt processing and meta tensor handling
        if enabled and "SDXL" in self.version and self.use_compel:
            self.logger.warning(
                "Disabling sequential CPU offload for SDXL model with "
                "compel due to compatibility issues"
            )
            enabled = False

        if enabled and not self.memory_settings.enable_model_cpu_offload:
            self._pipe.to("cpu")
            try:
                self.logger.debug("Enabling sequential cpu offload")
                self._pipe.enable_sequential_cpu_offload(self._device_index)
            except NotImplementedError as e:
                self.logger.warning(
                    f"Error applying sequential cpu offload: {e}"
                )
                self._pipe.to(self._device)
        else:
            self.logger.debug("Sequential cpu offload disabled")

    def _apply_model_offload(self, attr_val):
        """
        Apply model CPU offload for VRAM savings.

        Args:
            attr_val: Whether to enable model CPU offload.

        Less aggressive than sequential offload, keeps active components on GPU.
        May cause stability issues with SDXL + Compel.
        """
        enabled = AIRUNNER_MEM_ENABLE_MODEL_CPU_OFFLOAD
        if enabled is None:
            enabled = attr_val

        # Add warning for SDXL models with model CPU offload when using compel
        if enabled and "SDXL" in self.version and self.use_compel:
            self.logger.warning(
                "Model CPU offload with SDXL + compel may cause "
                "stability issues"
            )

        if (
            enabled
            and not self.memory_settings.use_enable_sequential_cpu_offload
        ):
            self.logger.debug("Enabling model cpu offload")
            # self._move_stable_diffusion_to_cpu()
            self._pipe.enable_model_cpu_offload(self._device_index)
        else:
            self.logger.debug("Model cpu offload disabled")

    def _apply_tome(self, attr_val):
        """
        Apply ToMe SD (Token Merging) for faster inference.

        Args:
            attr_val: Whether to enable ToMe SD.

        ToMe merges similar tokens to reduce computation. Minimal quality loss
        with significant speedup.
        """
        enabled = AIRUNNER_MEM_USE_TOME_SD
        if enabled is None:
            enabled = attr_val
        if enabled:
            ratio = AIRUNNER_MEM_TOME_SD_RATIO
            if ratio is None:
                ratio = self.memory_settings.tome_sd_ratio / 1000
            else:
                ratio = float(ratio)
            self.logger.debug(
                f"Applying ToMe SD weight merging with ratio {ratio}"
            )
            self._remove_tome_sd()
            try:
                tomesd.apply_patch(self._pipe, ratio=ratio)
            except Exception as e:
                self.logger.error(
                    f"Error applying ToMe SD weight merging: {e}"
                )
        else:
            self._remove_tome_sd()

    def _apply_xformers(self):
        """
        Apply xformers memory-efficient attention (Flash Attention 2).

        This provides significant speedup on modern GPUs (RTX 30xx/40xx).
        Can be disabled via AIRUNNER_DISABLE_FLASH_ATTENTION environment variable.
        """
        if AIRUNNER_DISABLE_FLASH_ATTENTION:
            self.logger.debug(
                "xformers disabled via AIRUNNER_DISABLE_FLASH_ATTENTION"
            )
            return

        if not is_ampere_or_newer(self._device_index):
            self.logger.debug(
                "xformers requires Ampere or newer GPU (RTX 30xx/40xx+)"
            )
            return

        try:
            if hasattr(
                self._pipe, "enable_xformers_memory_efficient_attention"
            ):
                self._pipe.enable_xformers_memory_efficient_attention()
                self.logger.info(
                    "✓ Enabled xformers memory-efficient attention (Flash Attention 2)"
                )
            else:
                self.logger.debug("Pipeline does not support xformers")
        except Exception as e:
            self.logger.warning(f"Could not enable xformers: {e}")
            self.logger.debug(
                "Install xformers for Flash Attention 2: pip install xformers"
            )

    def _apply_torch_compile(self):
        """
        Apply torch.compile() to UNet for 2-3x inference speedup.

        NOTE: torch.compile() is lazy - it wraps the model but doesn't compile
        until the first forward pass. This means:
        - Wrapping is instant (happens here)
        - Compilation is slow (happens during first generation, 2-3 minutes)
        - Cache persists per input shape (different resolutions = recompile)

        Only applies once per pipeline load to avoid recompilation overhead.

        Note: DeepCache has been permanently disabled as it's incompatible
        with torch.compile() and provides inferior speedup (15% vs 2-3x).
        """
        settings = get_qsettings()
        settings.beginGroup("generator_settings")
        enable_torch_compile = settings.value(
            "enable_torch_compile", False, type=bool
        )
        settings.endGroup()
        print("APPLY TORCH COMPILE")
        print("-" * 100)
        print(enable_torch_compile)
        if not enable_torch_compile:
            return

        if self._memory_settings_flags.get("torch_compile_applied"):
            return  # Already compiled

        if not hasattr(self._pipe, "unet") or self._pipe.unet is None:
            return

        try:
            self.logger.info(
                "Wrapping UNet with torch.compile() - compilation will happen on first generation"
            )
            # Note: torch.compile caches are stored in PyTorch's internal cache
            # directory (~/.triton or TORCH_COMPILE_DIR). They persist across runs
            # and are reused when the same model graph AND input shapes are seen.
            self._pipe.unet = torch.compile(
                self._pipe.unet,
                mode="reduce-overhead",  # Best for inference
                fullgraph=False,  # Allow fallback for unsupported ops
            )
            self._memory_settings_flags["torch_compile_applied"] = True
            self.logger.info(
                "✓ UNet wrapped for compilation (first generation will take 2-3 min)"
            )
            self.logger.debug(
                "Compiled cache stored in ~/.triton per input shape (resolution/batch)"
            )
        except Exception as e:
            self.logger.warning(f"Could not compile UNet: {e}")
            self.logger.debug("torch.compile requires PyTorch 2.0+")

    def _remove_tome_sd(self):
        """
        Remove ToMe SD patch from pipeline.

        Restores normal token processing.
        """
        self.logger.debug("Removing ToMe SD weight merging")
        try:
            tomesd.remove_patch(self._pipe)
        except Exception as e:
            self.logger.error(f"Error removing ToMe SD weight merging: {e}")

    def _clear_memory_efficient_settings(self):
        """
        Clear all memory efficiency flags.

        Forces reapplication of memory settings on next call.
        """
        self.logger.debug("Clearing memory efficient settings")
        for key in self._memory_settings_flags:
            if key.endswith("_applied"):
                self._memory_settings_flags[key] = None



"""Audio utilities for AI Runner.

This package contains utilities for managing audio devices and processing.
"""

from airunner.utils.audio.sound_device_manager import SoundDeviceManager

__all__ = ["SoundDeviceManager"]



"""
API Server Thread for headless mode.

Runs an HTTP server with /health, /llm, /art, /stt, /tts endpoints
without requiring Qt GUI components.
"""

import threading
from http.server import HTTPServer
from socketserver import ThreadingMixIn
from airunner.components.server.api.server import AIRunnerAPIRequestHandler
from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger
from airunner.settings import (
    AIRUNNER_HEADLESS_SERVER_HOST,
    AIRUNNER_HEADLESS_SERVER_PORT,
)

logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """Handle requests in a separate thread."""

    daemon_threads = True
    allow_reuse_address = True


class APIServerThread(threading.Thread):
    """Background thread running the AI Runner API server.

    Provides HTTP endpoints for LLM, art generation, TTS, and STT.
    Designed for headless operation without Qt GUI.

    Args:
        host: Host address to bind to (default: 0.0.0.0)
        port: Port to listen on (default: 8080)
    """

    def __init__(
        self,
        host: str = AIRUNNER_HEADLESS_SERVER_HOST,
        port: int = AIRUNNER_HEADLESS_SERVER_PORT,
    ):
        super().__init__(daemon=True)
        self.host = host
        self.port = port
        self.server = None
        self._stop_event = threading.Event()

    def run(self):
        """Start the HTTP server and serve requests."""
        try:
            self.server = ThreadedHTTPServer(
                (self.host, self.port), AIRunnerAPIRequestHandler
            )
            logger.info(
                f"API server listening on http://{self.host}:{self.port}"
            )
            logger.info("Available endpoints: /health, /llm, /art, /stt, /tts")

            # Serve requests until stopped
            while not self._stop_event.is_set():
                self.server.handle_request()

        except Exception as e:
            logger.error(f"API server error: {e}", exc_info=True)
        finally:
            if self.server:
                self.server.server_close()

    def stop(self):
        """Stop the server gracefully."""
        logger.info("Stopping API server...")
        self._stop_event.set()
        if self.server:
            self.server.shutdown()



from PySide6.QtCore import Qt
from PySide6.QtGui import QColor, QPainter, QPen, QBrush, QPixmap, QCursor


def circle_cursor(
    outline_color,
    fill_color,
    pixmap_size=32,
    *,
    pixmap_factory=QPixmap,
    painter_factory=QPainter,
    cursor_factory=QCursor
):
    """Create a custom circle cursor with dependency injection for testability.

    Args:
        outline_color: Color for the circle outline.
        fill_color: Color for the circle fill.
        pixmap_size: Size of the cursor pixmap.
        pixmap_factory: Factory for QPixmap (for testing).
        painter_factory: Factory for QPainter (for testing).
        cursor_factory: Factory for QCursor (for testing).
    Returns:
        QCursor: The created cursor.
    """
    pixmap_size = int(pixmap_size)
    pixmap = pixmap_factory(pixmap_size, pixmap_size)
    pixmap.fill(Qt.GlobalColor.transparent)
    painter = painter_factory(pixmap)
    painter.setRenderHint(QPainter.RenderHint.Antialiasing)
    pen = QPen(QColor(outline_color))
    pen.setWidth(2)
    brush = QBrush(QColor(fill_color))
    painter.setPen(pen)
    painter.setBrush(brush)
    painter.drawEllipse(1, 1, pixmap_size - 2, pixmap_size - 2)
    pen = QPen(QColor(Qt.GlobalColor.black))
    pen.setWidth(2)
    brush = QBrush(QColor(fill_color))
    painter.setPen(pen)
    painter.setBrush(brush)
    painter.drawEllipse(2, 2, pixmap_size - 4, pixmap_size - 4)
    painter.end()
    return cursor_factory(pixmap, pixmap_size // 2, pixmap_size // 2)



"""Download operations mixin for dataset manager."""

import os
import logging
from pathlib import Path
from typing import Optional, Callable, TYPE_CHECKING

import requests
from tqdm import tqdm

if TYPE_CHECKING:
    from ..dataset_manager import DatasetManager

logger = logging.getLogger(__name__)


class DatasetDownloadMixin:
    """Handles dataset downloading from HuggingFace."""

    def _construct_download_url(
        self: "DatasetManager", path: str, file_name: str
    ) -> str:
        """Construct HuggingFace dataset URL."""
        return (
            f"https://huggingface.co/datasets/{path}/resolve/main/{file_name}"
        )

    def _get_auth_headers(self: "DatasetManager") -> dict:
        """Get authentication headers with HF token if available."""
        token = (
            os.getenv("AIRUNNER_HF_TOKEN")
            or os.getenv("HF_TOKEN")
            or os.getenv("HUGGINGFACE_TOKEN")
        )
        if token:
            return {"Authorization": f"Bearer {token}"}
        return {}

    def _download_with_progress(
        self: "DatasetManager",
        response,
        dest_file: Path,
        file_name: str,
        callback: Optional[Callable[[int, int], None]] = None,
    ) -> None:
        """Download file with progress tracking."""
        total_size = int(response.headers.get("content-length", 0))

        if self.headless:
            self._download_with_tqdm(
                response, dest_file, file_name, total_size, callback
            )
        else:
            self._download_with_callback(
                response, dest_file, total_size, callback
            )

        logger.info(
            f"Successfully downloaded {dest_file} ({total_size} bytes)"
        )

    def _download_with_tqdm(
        self: "DatasetManager",
        response,
        dest_file: Path,
        file_name: str,
        total_size: int,
        callback: Optional[Callable],
    ) -> None:
        """Download with tqdm progress bar."""
        with tqdm(
            total=total_size,
            unit="B",
            unit_scale=True,
            desc=file_name.split("/")[-1],
        ) as pbar:
            with open(dest_file, "wb") as f:
                downloaded = 0
                for chunk in response.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
                        downloaded += len(chunk)
                        pbar.update(len(chunk))
                        if callback:
                            callback(downloaded, total_size)

    def _download_with_callback(
        self: "DatasetManager",
        response,
        dest_file: Path,
        total_size: int,
        callback: Optional[Callable],
    ) -> None:
        """Download with callback only."""
        with open(dest_file, "wb") as f:
            downloaded = 0
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)
                    downloaded += len(chunk)
                    if callback:
                        callback(downloaded, total_size)

    def _download_file(
        self: "DatasetManager",
        path: str,
        file_name: str,
        file_path: str,
        callback: Optional[Callable[[int, int], None]] = None,
    ) -> None:
        """Download a file directly using requests."""
        url = self._construct_download_url(path, file_name)
        dest_dir = Path(file_path)
        dest_dir.mkdir(parents=True, exist_ok=True)
        dest_file = dest_dir / file_name.split("/")[-1]

        logger.info(f"Downloading {url} to {dest_file}")

        headers = self._get_auth_headers()

        try:
            response = self._make_download_request(url, headers)
            self._download_with_progress(
                response, dest_file, file_name, callback
            )

        except requests.exceptions.HTTPError as e:
            self._handle_http_error(e)
            raise
        except Exception as e:
            logger.error(f"Download failed for {url}: {e}")
            raise

    def _make_download_request(
        self: "DatasetManager", url: str, headers: dict
    ):
        """Make HTTP request for download."""
        response = requests.get(
            url,
            stream=True,
            headers=headers,
            timeout=60,
        )
        response.raise_for_status()
        return response

    def _handle_http_error(
        self: "DatasetManager", error: requests.exceptions.HTTPError
    ) -> None:
        """Handle HTTP errors with helpful messages."""
        if error.response.status_code == 401:
            logger.error(
                "Unauthorized (401). A Hugging Face token may be "
                "required. Set AIRUNNER_HF_TOKEN, HF_TOKEN, or "
                "HUGGINGFACE_TOKEN environment variable."
            )

    def download_gsm8k(
        self: "DatasetManager",
        split: str = "test",
        progress_callback: Optional[Callable[[int, int], None]] = None,
    ) -> Path:
        """Download GSM8K dataset."""
        config = self.DATASETS["gsm8k"]
        repo = config["repo"]
        filename = config["files"][split]

        dest_path = self.get_dataset_path("gsm8k", split)
        dest_dir = dest_path.parent
        dest_dir.mkdir(parents=True, exist_ok=True)

        if dest_path.exists():
            logger.info(f"GSM8K {split} already exists at {dest_path}")
            return dest_path

        logger.info(f"Downloading GSM8K {split} from {repo}")
        self._download_file(repo, filename, str(dest_dir), progress_callback)

        return dest_path

    def download_math(
        self: "DatasetManager",
        split: str = "test",
        subjects: Optional[list] = None,
        progress_callback: Optional[Callable[[int, int], None]] = None,
    ) -> Path:
        """Download MATH dataset."""
        config = self.DATASETS["math"]
        repo = config["repo"]

        if subjects is None:
            subjects = config["subjects"]

        dest_dir = self.get_dataset_path("math", split)
        dest_dir.mkdir(parents=True, exist_ok=True)

        logger.info(f"Downloading MATH {split} for {len(subjects)} subjects")

        self._download_subjects(
            repo, split, subjects, dest_dir, progress_callback
        )

        return dest_dir

    def _download_subjects(
        self: "DatasetManager",
        repo: str,
        split: str,
        subjects: list,
        dest_dir: Path,
        progress_callback: Optional[Callable],
    ) -> None:
        """Download all subjects for MATH dataset."""
        total_subjects = len(subjects)

        def wrapped_callback(current: int, total: int):
            """Wrapper callback for progress tracking.

            Args:
                current: Current bytes downloaded
                total: Total bytes to download
            """
            if progress_callback:
                progress_callback(current, total)

        for idx, subject in enumerate(subjects):
            self._download_subject(
                repo,
                split,
                subject,
                dest_dir,
                idx,
                total_subjects,
                wrapped_callback,
            )

    def _download_subject(
        self: "DatasetManager",
        repo: str,
        split: str,
        subject: str,
        dest_dir: Path,
        idx: int,
        total: int,
        callback: Callable,
    ) -> None:
        """Download a single subject parquet file."""
        filename = f"{split}/{subject}/default-00000-of-00001.parquet"
        subject_file = dest_dir / f"{subject}.parquet"

        if subject_file.exists():
            logger.info(
                f"Subject {subject} ({idx + 1}/{total}) already exists"
            )
            return

        logger.info(f"Downloading subject {subject} ({idx + 1}/{total})")
        self._download_file(repo, filename, str(dest_dir), callback)



"""RAG Search Node for LangGraph workflows.

This node performs RAG search using LlamaIndex.
"""

from typing import List
from airunner.vendor.nodegraphqt.constants import NodePropWidgetEnum
from airunner.components.nodegraph.gui.widgets.nodes.langgraph.base_langgraph_node import (
    BaseLangGraphNode,
)


class RAGSearchNode(BaseLangGraphNode):
    """Perform RAG search using LlamaIndex.

    This node retrieves relevant documents from a vector store
    and adds them to the workflow state.
    """

    NODE_NAME = "RAG Search"
    state_key = "rag_context"

    _input_ports = [
        dict(name="query", display_name="Query"),
    ]

    _output_ports = [
        dict(name="context", display_name="Context"),
        dict(name="documents", display_name="Documents"),
    ]

    _properties = [
        dict(
            name="top_k",
            value=5,
            widget_type=NodePropWidgetEnum.QSPIN_BOX,
            range=(1, 20),
            tab="search",
        ),
        dict(
            name="query_key",
            value="query",
            widget_type=NodePropWidgetEnum.QLINE_EDIT,
            tab="state",
        ),
        dict(
            name="context_key",
            value="rag_context",
            widget_type=NodePropWidgetEnum.QLINE_EDIT,
            tab="state",
        ),
        dict(
            name="docs_key",
            value="retrieved_docs",
            widget_type=NodePropWidgetEnum.QLINE_EDIT,
            tab="state",
        ),
    ]

    def get_node_type(self) -> str:
        """Get node type identifier."""
        return "rag"

    def get_description(self) -> str:
        """Get node description."""
        top_k = self.get_property("top_k")
        return f"RAG search (top {top_k})"

    def to_langgraph_code(self) -> str:
        """Generate Python code for RAG search.

        Returns:
            Python code string
        """
        top_k = self.get_property("top_k")
        query_key = self.get_property("query_key")
        context_key = self.get_property("context_key")
        docs_key = self.get_property("docs_key")

        func_name = self._sanitize_name(self.name())

        code = f'''def {func_name}(state: AgentState) -> AgentState:
    """{self.get_description()}"""
    # NOTE: RAG index must be provided externally
    # This is a bridge to existing LlamaIndex functionality
    
    query = state.get("{query_key}", "")
    if not query:
        # Try getting from messages
        messages = state.get("messages", [])
        query = messages[-1] if messages else ""
    
    if not query:
        logger.warning("No query found for RAG search")
        return state
    
    try:
        # Use LlamaIndex bridge (must be configured)
        from airunner.components.llm.langgraph.bridge import LlamaIndexBridge
        
        # This assumes a global rag_index is available
        # In practice, this would be passed via config
        if hasattr(state.get("metadata", {{}}), "rag_index"):
            rag_index = state["metadata"]["rag_index"]
            retriever = rag_index.as_retriever(similarity_top_k={top_k})
            nodes = retriever.retrieve(query)
            
            context_parts = [node.text for node in nodes]
            doc_metadata = [
                {{"text": node.text[:200], "score": node.score}}
                for node in nodes
            ]
            
            state["{context_key}"] = "\\n\\n".join(context_parts)
            state["{docs_key}"] = doc_metadata
            logger.info(f"Retrieved {{len(nodes)}} documents")
        else:
            logger.warning("No RAG index found in state metadata")
    
    except Exception as e:
        logger.error(f"RAG search error: {{e}}")
        state["error"] = str(e)
    
    return state'''

        return code

    def get_input_state_keys(self) -> List[str]:
        """Get state keys this node reads from."""
        return [self.get_property("query_key"), "messages"]

    def get_output_state_keys(self) -> List[str]:
        """Get state keys this node writes to."""
        return [
            self.get_property("context_key"),
            self.get_property("docs_key"),
        ]

    @staticmethod
    def _sanitize_name(name: str) -> str:
        """Sanitize node name to valid Python identifier."""
        sanitized = "".join(
            c if c.isalnum() or c == "_" else "_" for c in name
        )
        if sanitized and sanitized[0].isdigit():
            sanitized = f"node_{sanitized}"
        return sanitized or "rag_node"



from __future__ import annotations

from typing import Optional, Set, TYPE_CHECKING

from PySide6.QtCore import QPointF, QRectF, Qt
from PySide6.QtGui import QColor, QPainter, QPen
from PySide6.QtWidgets import (
    QGraphicsRectItem,
    QGraphicsTextItem,
    QMenu,
)

from airunner.enums import CanvasToolName
from airunner.components.art.utils.canvas_position_manager import (
    CanvasPositionManager,
)

if TYPE_CHECKING:
    from airunner.components.art.gui.widgets.canvas.custom_view import (
        CustomGraphicsView,
    )
else:  # pragma: no cover - avoid hard dependency during import cycles
    CustomGraphicsView = object


class ResizableTextItem(QGraphicsRectItem):
    """Rectangular text area that keeps text clipped and supports resizing."""

    MIN_WIDTH = 40.0
    MIN_HEIGHT = 40.0
    BORDER_MARGIN = 8.0

    _ANCHOR_TO_EDGES = {
        "left": {"left"},
        "right": {"right"},
        "top": {"top"},
        "bottom": {"bottom"},
        "top-left": {"top", "left"},
        "top-right": {"top", "right"},
        "bottom-left": {"bottom", "left"},
        "bottom-right": {"bottom", "right"},
    }

    def __init__(self, view: CustomGraphicsView, rect: QRectF):
        width = max(rect.width(), self.MIN_WIDTH)
        height = max(rect.height(), self.MIN_HEIGHT)
        super().__init__(QRectF(0.0, 0.0, width, height))

        self._view = view
        self.setPos(rect.topLeft())
        self.setZValue(2000)
        self.setPen(QPen(QColor("white")))
        self.setFlag(QGraphicsRectItem.GraphicsItemFlag.ItemIsSelectable, True)
        self.setFlag(QGraphicsRectItem.GraphicsItemFlag.ItemIsMovable, True)
        self.setFlag(
            QGraphicsRectItem.GraphicsItemFlag.ItemSendsGeometryChanges, True
        )
        self.setFlag(
            QGraphicsRectItem.GraphicsItemFlag.ItemClipsChildrenToShape, True
        )
        self.setAcceptHoverEvents(True)

        self.text_item = QGraphicsTextItem(self)
        self.text_item.setDefaultTextColor(QColor("white"))
        try:
            self.text_item.setFont(self._view._get_default_text_font())
        except Exception:
            pass
        self.text_item.setTextInteractionFlags(Qt.TextEditorInteraction)
        self.text_item.setPos(QPointF(0.0, 0.0))
        self.text_item.setTextWidth(width)
        self.text_item.setZValue(self.zValue() + 1)

        self._resizing = False
        self._resize_anchor: Optional[str] = None
        self._initial_geometry: tuple[float, float, float, float] = (
            rect.left(),
            rect.top(),
            rect.left() + width,
            rect.top() + height,
        )
        self._start_scene_pos = QPointF(rect.topLeft())
        self._initial_aspect = width / height if height > 0 else None

    @property
    def current_tool(self) -> Optional[CanvasToolName]:
        try:
            return self._view.current_tool
        except Exception:
            return None

    def setText(self, text: str) -> None:
        self.text_item.setPlainText(text)

    def set_interaction_enabled(self, enable: bool) -> None:
        self.setFlag(QGraphicsRectItem.GraphicsItemFlag.ItemIsMovable, enable)
        self.setFlag(
            QGraphicsRectItem.GraphicsItemFlag.ItemIsSelectable, enable
        )
        self.text_item.setTextInteractionFlags(
            Qt.TextEditorInteraction if enable else Qt.NoTextInteraction
        )

    def hoverMoveEvent(self, event):
        anchor = self._detect_anchor(self.mapFromScene(event.scenePos()))
        cursor = self._cursor_for_anchor(anchor)
        if cursor is not None:
            self.setCursor(cursor)
        else:
            self.unsetCursor()
        super().hoverMoveEvent(event)

    def hoverLeaveEvent(self, event):
        self.unsetCursor()
        super().hoverLeaveEvent(event)

    def mousePressEvent(self, event):
        if event.button() == Qt.MouseButton.RightButton:
            self._open_delete_menu(event)
            return

        if event.button() != Qt.MouseButton.LeftButton:
            super().mousePressEvent(event)
            return

        anchor = self._detect_anchor(self.mapFromScene(event.scenePos()))
        if anchor is not None:
            self._begin_resize(anchor, event.scenePos())
            event.accept()
            return

        if self.current_tool != CanvasToolName.MOVE:
            self.text_item.setFocus()
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if self._resizing and self._resize_anchor:
            self._apply_resize(event.scenePos(), event.modifiers())
            event.accept()
            return
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if self._resizing:
            self._finish_resize(event.scenePos())
            event.accept()
            return
        super().mouseReleaseEvent(event)
        if event.button() == Qt.MouseButton.LeftButton:
            self._persist_geometry()

    def itemChange(self, change, value):
        if change == QGraphicsRectItem.GraphicsItemChange.ItemPositionChange:
            value = self._snap_position_if_needed(value)
        elif (
            change
            == QGraphicsRectItem.GraphicsItemChange.ItemPositionHasChanged
        ):
            self._persist_geometry()
        return super().itemChange(change, value)

    def paint(self, painter: QPainter, option, widget=None):
        super().paint(painter, option, widget)
        if not self.isSelected():
            return
        size = 6.0
        rect = self.rect()
        painter.save()
        painter.setPen(QPen(QColor("white")))
        painter.setBrush(QColor("white"))
        handle_rects = [
            QRectF(rect.left() - size / 2, rect.top() - size / 2, size, size),
            QRectF(rect.right() - size / 2, rect.top() - size / 2, size, size),
            QRectF(
                rect.left() - size / 2, rect.bottom() - size / 2, size, size
            ),
            QRectF(
                rect.right() - size / 2, rect.bottom() - size / 2, size, size
            ),
        ]
        for handle_rect in handle_rects:
            painter.drawRect(handle_rect)
        painter.restore()

    def to_persist_dict(self) -> dict:
        """Return a JSON-serializable representation of this item."""

        abs_x = int(self.scenePos().x() + self._view.canvas_offset_x)
        abs_y = int(self.scenePos().y() + self._view.canvas_offset_y)
        return {
            "type": "area",
            "text": self.text_item.toPlainText(),
            "x": abs_x,
            "y": abs_y,
            "color": self.text_item.defaultTextColor().name(),
            "font": self.text_item.font().toString(),
            "w": int(self.rect().width()),
            "h": int(self.rect().height()),
        }

    def _persist_geometry(self) -> None:
        if self._view is not None:
            try:
                self._view._save_text_items_to_db()
            except Exception:
                pass

    def _begin_resize(self, anchor: str, scene_pos: QPointF) -> None:
        self._resizing = True
        self._resize_anchor = anchor
        self._start_scene_pos = QPointF(scene_pos)
        self._capture_initial_geometry()
        self.text_item.setTextInteractionFlags(Qt.NoTextInteraction)

    def _apply_resize(
        self, scene_pos: QPointF, modifiers: Qt.KeyboardModifiers
    ) -> None:
        if not self._resizing or self._resize_anchor is None:
            return

        edges = self._ANCHOR_TO_EDGES[self._resize_anchor]
        delta = scene_pos - self._start_scene_pos
        left, top, right, bottom = self._initial_geometry

        if "left" in edges:
            left += delta.x()
        if "right" in edges:
            right += delta.x()
        if "top" in edges:
            top += delta.y()
        if "bottom" in edges:
            bottom += delta.y()

        left, right = self._enforce_min_width(edges, left, right)
        top, bottom = self._enforce_min_height(edges, top, bottom)

        if modifiers & Qt.KeyboardModifier.ShiftModifier:
            left, right, top, bottom = self._apply_aspect_constraint(
                edges, left, right, top, bottom
            )

        left, right, top, bottom = self._apply_grid_snapping(
            edges, left, right, top, bottom
        )

        width = max(self.MIN_WIDTH, right - left)
        height = max(self.MIN_HEIGHT, bottom - top)

        self.prepareGeometryChange()
        self.setRect(QRectF(0.0, 0.0, width, height))
        self.setPos(QPointF(left, top))
        self.text_item.setTextWidth(width)

    def _finish_resize(self, scene_pos: QPointF) -> None:
        self._apply_resize(scene_pos, Qt.KeyboardModifier.NoModifier)
        self._resizing = False
        self._resize_anchor = None
        self.text_item.setTextInteractionFlags(Qt.TextEditorInteraction)
        self._persist_geometry()

    def _capture_initial_geometry(self) -> None:
        rect = self.rect()
        pos = self.scenePos()
        self._initial_geometry = (
            pos.x(),
            pos.y(),
            pos.x() + rect.width(),
            pos.y() + rect.height(),
        )
        height = rect.height()
        self._initial_aspect = rect.width() / height if height > 0 else None

    def _apply_aspect_constraint(
        self,
        edges: Set[str],
        left: float,
        right: float,
        top: float,
        bottom: float,
    ) -> tuple[float, float, float, float]:
        if not self._initial_aspect:
            return left, right, top, bottom

        width = max(self.MIN_WIDTH, right - left)
        height = max(self.MIN_HEIGHT, bottom - top)
        target_height = width / self._initial_aspect
        target_width = height * self._initial_aspect
        anchor = self._resize_anchor or ""

        if anchor in {"left", "right"}:
            height = target_height
            if "top" in edges and "bottom" not in edges:
                top = bottom - height
            else:
                bottom = top + height
        elif anchor in {"top", "bottom"}:
            width = target_width
            if "left" in edges and "right" not in edges:
                left = right - width
            else:
                right = left + width
        else:
            width_change = abs(
                width - (self._initial_geometry[2] - self._initial_geometry[0])
            )
            height_change = abs(
                height
                - (self._initial_geometry[3] - self._initial_geometry[1])
            )
            if width_change >= height_change:
                height = target_height
                if "top" in edges and "bottom" not in edges:
                    top = bottom - height
                else:
                    bottom = top + height
            else:
                width = target_width
                if "left" in edges and "right" not in edges:
                    left = right - width
                else:
                    right = left + width

        return left, right, top, bottom

    def _apply_grid_snapping(
        self,
        edges: Set[str],
        left: float,
        right: float,
        top: float,
        bottom: float,
    ) -> tuple[float, float, float, float]:
        grid = getattr(self._view, "grid_settings", None)
        if not grid or not getattr(grid, "snap_to_grid", False):
            return left, right, top, bottom

        manager = CanvasPositionManager()
        offset_x = self._view.canvas_offset_x
        offset_y = self._view.canvas_offset_y

        if "left" in edges or "top" in edges:
            abs_pos = QPointF(left + offset_x, top + offset_y)
            snapped = manager.snap_to_grid(abs_pos, grid.cell_size)
            if "left" in edges:
                left = snapped.x() - offset_x
            if "top" in edges:
                top = snapped.y() - offset_y

        if "right" in edges or "bottom" in edges:
            abs_pos = QPointF(right + offset_x, bottom + offset_y)
            snapped = manager.snap_to_grid(abs_pos, grid.cell_size)
            if "right" in edges:
                right = snapped.x() - offset_x
            if "bottom" in edges:
                bottom = snapped.y() - offset_y

        return left, right, top, bottom

    def _enforce_min_width(
        self, edges: Set[str], left: float, right: float
    ) -> tuple[float, float]:
        width = right - left
        if width >= self.MIN_WIDTH:
            return left, right
        if "left" in edges and "right" not in edges:
            left = right - self.MIN_WIDTH
        else:
            right = left + self.MIN_WIDTH
        return left, right

    def _enforce_min_height(
        self, edges: Set[str], top: float, bottom: float
    ) -> tuple[float, float]:
        height = bottom - top
        if height >= self.MIN_HEIGHT:
            return top, bottom
        if "top" in edges and "bottom" not in edges:
            top = bottom - self.MIN_HEIGHT
        else:
            bottom = top + self.MIN_HEIGHT
        return top, bottom

    def _open_delete_menu(self, event) -> None:
        menu = QMenu()
        delete_action = menu.addAction(self.tr("Delete"))
        chosen = menu.exec_(event.screenPos())
        if chosen != delete_action:
            return

        view = getattr(self, "_view", None)
        if view is not None:
            try:
                view._remove_text_item(self)
            except Exception:
                pass

    def _detect_anchor(self, local_pos: QPointF) -> Optional[str]:
        rect = self.rect()
        margin = self.BORDER_MARGIN
        left = abs(local_pos.x() - rect.left()) <= margin
        right = abs(local_pos.x() - rect.right()) <= margin
        top = abs(local_pos.y() - rect.top()) <= margin
        bottom = abs(local_pos.y() - rect.bottom()) <= margin

        if left and top:
            return "top-left"
        if right and top:
            return "top-right"
        if left and bottom:
            return "bottom-left"
        if right and bottom:
            return "bottom-right"
        if left:
            return "left"
        if right:
            return "right"
        if top:
            return "top"
        if bottom:
            return "bottom"
        return None

    @staticmethod
    def _cursor_for_anchor(anchor: Optional[str]):
        if anchor in {"top-left", "bottom-right"}:
            return Qt.CursorShape.SizeFDiagCursor
        if anchor in {"top-right", "bottom-left"}:
            return Qt.CursorShape.SizeBDiagCursor
        if anchor in {"left", "right"}:
            return Qt.CursorShape.SizeHorCursor
        if anchor in {"top", "bottom"}:
            return Qt.CursorShape.SizeVerCursor
        return None

    def _snap_position_if_needed(self, value):
        if not isinstance(value, QPointF):
            return value
        if self.current_tool != CanvasToolName.MOVE:
            return value

        grid = getattr(self._view, "grid_settings", None)
        if not grid or not getattr(grid, "snap_to_grid", False):
            return value

        manager = CanvasPositionManager()
        abs_pos = QPointF(
            value.x() + self._view.canvas_offset_x,
            value.y() + self._view.canvas_offset_y,
        )
        snapped = manager.snap_to_grid(abs_pos, grid.cell_size)
        return QPointF(
            snapped.x() - self._view.canvas_offset_x,
            snapped.y() - self._view.canvas_offset_y,
        )



"""Project manager for long-running agent projects.

Handles CRUD operations for projects, features, progress entries,
and session state. Acts as the persistence layer for the harness.
"""

from datetime import datetime
from typing import Optional, List, Dict, Any
from pathlib import Path
import json
import subprocess

from sqlalchemy.orm import Session, make_transient

from airunner.components.llm.long_running.data.project_state import (
    ProjectState,
    ProjectFeature,
    ProgressEntry,
    SessionState,
    DecisionMemory,
    ProjectStatus,
    FeatureStatus,
    FeatureCategory,
    DecisionOutcome,
)
from airunner.components.data.session_manager import session_scope
from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger

logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)


class ProjectManager:
    """Manages long-running agent projects.

    Provides methods for creating projects, managing features,
    logging progress, and maintaining session state.

    Example:
        ```python
        manager = ProjectManager()

        # Create a new project
        project = manager.create_project(
            name="My Web App",
            description="Build a chat application clone",
            working_directory="/path/to/project"
        )

        # Add features
        manager.add_feature(
            project_id=project.id,
            name="User can send messages",
            description="User types a message and it appears in the chat",
            category=FeatureCategory.FUNCTIONAL,
            verification_steps=["Open chat", "Type message", "Click send", "See message"]
        )

        # Start a session
        session = manager.start_session(project.id)

        # Log progress
        manager.log_progress(
            project_id=project.id,
            session_id=session.id,
            action="Implemented message sending",
            outcome="Messages now appear in chat"
        )

        # Mark feature as passing
        manager.update_feature_status(feature_id, FeatureStatus.PASSING)

        # End session
        manager.end_session(session.id, next_action="Add message timestamps")
        ```
    """

    def __init__(self):
        """Initialize project manager."""
        self._logger = logger

    def _detach(self, db: Session, obj):
        """Detach an object from the session so it can be used outside.
        
        Args:
            db: The database session
            obj: The ORM object to detach
            
        Returns:
            The detached object
        """
        if obj is not None:
            db.expunge(obj)
            make_transient(obj)
        return obj
    
    def _detach_all(self, db: Session, objects: List):
        """Detach a list of objects from the session.
        
        Args:
            db: The database session
            objects: List of ORM objects to detach
            
        Returns:
            List of detached objects
        """
        for obj in objects:
            self._detach(db, obj)
        return objects

    # =========================================================================
    # Project Operations
    # =========================================================================

    def create_project(
        self,
        name: str,
        description: str,
        working_directory: Optional[str] = None,
        system_prompt: Optional[str] = None,
        init_git: bool = True,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> ProjectState:
        """Create a new long-running project.

        Args:
            name: Human-readable project name
            description: What the project aims to achieve
            working_directory: Filesystem path for project files
            system_prompt: Base system prompt for agents
            init_git: Whether to initialize git repository
            metadata: Additional metadata

        Returns:
            Created ProjectState instance

        Raises:
            ValueError: If project with name already exists
        """
        with session_scope() as db:
            # Check for existing project
            existing = (
                db.query(ProjectState)
                .filter(ProjectState.name == name)
                .first()
            )
            if existing:
                raise ValueError(f"Project '{name}' already exists")

            # Create working directory if specified
            if working_directory:
                work_dir = Path(working_directory)
                work_dir.mkdir(parents=True, exist_ok=True)

                # Initialize git if requested
                git_path = None
                if init_git:
                    git_path = self._init_git_repo(work_dir)

            project = ProjectState(
                name=name,
                description=description,
                working_directory=working_directory,
                git_repo_path=git_path if working_directory and init_git else None,
                status=ProjectStatus.INITIALIZING,
                system_prompt=system_prompt,
                project_metadata=metadata or {},
            )
            db.add(project)
            db.commit()
            db.refresh(project)
            
            self._logger.info(f"Created project '{name}' (ID: {project.id})")
            return self._detach(db, project)

    def get_project(self, project_id: int) -> Optional[ProjectState]:
        """Get project by ID.

        Args:
            project_id: Project ID

        Returns:
            ProjectState or None if not found
        """
        with session_scope() as db:
            project = db.query(ProjectState).filter(ProjectState.id == project_id).first()
            return self._detach(db, project)

    def get_project_by_name(self, name: str) -> Optional[ProjectState]:
        """Get project by name.

        Args:
            name: Project name

        Returns:
            ProjectState or None if not found
        """
        with session_scope() as db:
            project = (
                db.query(ProjectState)
                .filter(ProjectState.name == name)
                .first()
            )
            return self._detach(db, project)

    def list_projects(
        self, status: Optional[ProjectStatus] = None
    ) -> List[ProjectState]:
        """List all projects, optionally filtered by status.

        Args:
            status: Optional status filter

        Returns:
            List of ProjectState instances
        """
        with session_scope() as db:
            query = db.query(ProjectState)
            if status:
                query = query.filter(ProjectState.status == status)
            projects = query.order_by(ProjectState.updated_at.desc()).all()
            return self._detach_all(db, projects)

    def update_project_status(
        self, project_id: int, status: ProjectStatus
    ) -> None:
        """Update project status.

        Args:
            project_id: Project ID
            status: New status
        """
        with session_scope() as db:
            project = db.query(ProjectState).filter(ProjectState.id == project_id).first()
            if project:
                project.status = status
                project.updated_at = datetime.utcnow()
                db.commit()
                self._logger.info(
                    f"Project {project_id} status updated to {status.value}"
                )

    def delete_project(self, project_id: int) -> bool:
        """Delete a project and all related data.

        Args:
            project_id: Project ID

        Returns:
            True if deleted, False if not found
        """
        with session_scope() as db:
            project = db.query(ProjectState).filter(ProjectState.id == project_id).first()
            if not project:
                return False
            db.delete(project)
            db.commit()
            self._logger.info(f"Deleted project {project_id}")
            return True

    # =========================================================================
    # Feature Operations
    # =========================================================================

    def add_feature(
        self,
        project_id: int,
        name: str,
        description: str,
        category: FeatureCategory = FeatureCategory.FUNCTIONAL,
        verification_steps: Optional[List[str]] = None,
        priority: int = 5,
        depends_on: Optional[List[int]] = None,
    ) -> ProjectFeature:
        """Add a feature to a project.

        Args:
            project_id: Parent project ID
            name: Short feature name
            description: Detailed description
            category: Feature category
            verification_steps: Steps to verify feature works
            priority: 1-10 priority
            depends_on: IDs of features this depends on

        Returns:
            Created ProjectFeature instance
        """
        with session_scope() as db:
            feature = ProjectFeature(
                project_id=project_id,
                name=name,
                description=description,
                category=category,
                verification_steps=verification_steps or [],
                priority=priority,
                depends_on=depends_on or [],
                status=FeatureStatus.NOT_STARTED,
            )
            db.add(feature)

            # Update project feature count
            project = db.query(ProjectState).filter(ProjectState.id == project_id).first()
            if project:
                project.total_features = (project.total_features or 0) + 1

            db.commit()
            db.refresh(feature)

            self._logger.info(f"Added feature '{name}' to project {project_id}")
            return self._detach(db, feature)

    def add_features_bulk(
        self, project_id: int, features: List[Dict[str, Any]]
    ) -> List[ProjectFeature]:
        """Add multiple features at once.

        Args:
            project_id: Parent project ID
            features: List of feature dicts with keys:
                - name (required)
                - description (required)
                - category (optional)
                - verification_steps (optional)
                - priority (optional)
                - depends_on (optional)

        Returns:
            List of created features
        """
        created = []
        with session_scope() as db:
            for f in features:
                feature = ProjectFeature(
                    project_id=project_id,
                    name=f["name"],
                    description=f.get("description", ""),
                    category=FeatureCategory(f.get("category", "functional")),
                    verification_steps=f.get("verification_steps", []),
                    priority=f.get("priority", 5),
                    depends_on=f.get("depends_on", []),
                    status=FeatureStatus.NOT_STARTED,
                )
                db.add(feature)
                created.append(feature)

            # Update project feature count
            project = db.query(ProjectState).filter(ProjectState.id == project_id).first()
            if project:
                project.total_features = (project.total_features or 0) + len(features)

            db.commit()
            for f in created:
                db.refresh(f)

            self._logger.info(f"Added {len(created)} features to project {project_id}")
            return self._detach_all(db, created)

    def get_feature(self, feature_id: int) -> Optional[ProjectFeature]:
        """Get feature by ID.

        Args:
            feature_id: Feature ID

        Returns:
            ProjectFeature or None
        """
        with session_scope() as db:
            feature = (
                db.query(ProjectFeature)
                .filter(ProjectFeature.id == feature_id)
                .first()
            )
            return self._detach(db, feature)

    def get_project_features(
        self,
        project_id: int,
        status: Optional[FeatureStatus] = None,
        category: Optional[FeatureCategory] = None,
    ) -> List[ProjectFeature]:
        """Get features for a project with optional filters.

        Args:
            project_id: Project ID
            status: Optional status filter
            category: Optional category filter

        Returns:
            List of features
        """
        with session_scope() as db:
            query = db.query(ProjectFeature).filter(
                ProjectFeature.project_id == project_id
            )
            if status:
                query = query.filter(ProjectFeature.status == status)
            if category:
                query = query.filter(ProjectFeature.category == category)
            features = query.order_by(ProjectFeature.priority.desc()).all()
            return self._detach_all(db, features)

    def get_next_feature_to_work_on(
        self, project_id: int
    ) -> Optional[ProjectFeature]:
        """Get the next feature that should be worked on.

        Prioritizes by:
        1. In-progress features (resume)
        2. Not-started features with met dependencies
        3. Failing features (retry)

        Args:
            project_id: Project ID

        Returns:
            Next feature to work on, or None if all passing
        """
        with session_scope() as db:
            # First check for in-progress features
            in_progress = (
                db.query(ProjectFeature)
                .filter(
                    ProjectFeature.project_id == project_id,
                    ProjectFeature.status == FeatureStatus.IN_PROGRESS,
                )
                .first()
            )
            if in_progress:
                return self._detach(db, in_progress)

            # Get all features for dependency checking
            all_features = (
                db.query(ProjectFeature)
                .filter(ProjectFeature.project_id == project_id)
                .all()
            )
            passing_ids = {
                f.id for f in all_features if f.status == FeatureStatus.PASSING
            }

            # Find not-started features with met dependencies
            not_started = [
                f
                for f in all_features
                if f.status == FeatureStatus.NOT_STARTED
            ]
            for feature in sorted(
                not_started, key=lambda x: x.priority, reverse=True
            ):
                # Check dependencies
                deps = feature.depends_on or []
                if all(dep_id in passing_ids for dep_id in deps):
                    return self._detach(db, feature)

            # Fall back to failing features (retry)
            failing = [
                f
                for f in all_features
                if f.status == FeatureStatus.FAILING
            ]
            if failing:
                # Return the one with fewest attempts
                feature = min(failing, key=lambda x: x.attempts or 0)
                return self._detach(db, feature)

            return None

    def update_feature_status(
        self,
        feature_id: int,
        status: FeatureStatus,
        error: Optional[str] = None,
    ) -> None:
        """Update feature status.

        Args:
            feature_id: Feature ID
            status: New status
            error: Optional error message (for failing status)
        """
        with session_scope() as db:
            feature = (
                db.query(ProjectFeature)
                .filter(ProjectFeature.id == feature_id)
                .first()
            )
            if not feature:
                return

            old_status = feature.status
            feature.status = status
            feature.updated_at = datetime.utcnow()

            if status == FeatureStatus.FAILING:
                feature.attempts = (feature.attempts or 0) + 1
                feature.last_error = error
            elif status == FeatureStatus.PASSING:
                feature.last_error = None

            # Update project passing count
            project = (
                db.query(ProjectState)
                .filter(ProjectState.id == feature.project_id)
                .first()
            )
            if project:
                if (
                    old_status != FeatureStatus.PASSING
                    and status == FeatureStatus.PASSING
                ):
                    project.passing_features = (project.passing_features or 0) + 1
                elif (
                    old_status == FeatureStatus.PASSING
                    and status != FeatureStatus.PASSING
                ):
                    project.passing_features = max(
                        0, (project.passing_features or 0) - 1
                    )

                # Check if all features passing
                if (
                    project.passing_features == project.total_features
                    and project.total_features > 0
                ):
                    project.status = ProjectStatus.COMPLETED
                    self._logger.info(
                        f"Project {project.id} completed! All features passing."
                    )

            db.commit()
            self._logger.info(f"Feature {feature_id} status: {status.value}")

    # =========================================================================
    # Session Operations
    # =========================================================================

    def start_session(
        self,
        project_id: int,
        feature_id: Optional[int] = None,
        context_snapshot: Optional[Dict[str, Any]] = None,
    ) -> SessionState:
        """Start a new working session.

        Args:
            project_id: Project ID
            feature_id: Feature to work on (or None to auto-select)
            context_snapshot: Optional context to save

        Returns:
            Created SessionState
        """
        with session_scope() as db:
            # Auto-select feature if not specified
            if feature_id is None:
                feature = self.get_next_feature_to_work_on(project_id)
                feature_id = feature.id if feature else None

            # Mark feature as in-progress
            if feature_id:
                feat = (
                    db.query(ProjectFeature)
                    .filter(ProjectFeature.id == feature_id)
                    .first()
                )
                if feat:
                    feat.status = FeatureStatus.IN_PROGRESS

            session = SessionState(
                project_id=project_id,
                feature_id=feature_id,
                context_snapshot=context_snapshot or {},
                started_at=datetime.utcnow(),
            )
            db.add(session)

            # Update project current feature
            project = db.query(ProjectState).filter(ProjectState.id == project_id).first()
            if project:
                project.current_feature_id = feature_id

            db.commit()
            db.refresh(session)

            self._logger.info(
                f"Started session {session.id} for project {project_id}, "
                f"feature {feature_id}"
            )
            return self._detach(db, session)

    def end_session(
        self,
        session_id: int,
        working_memory: Optional[Dict[str, Any]] = None,
        next_action: Optional[str] = None,
        error: Optional[str] = None,
        tokens_consumed: int = 0,
    ) -> None:
        """End a working session.

        Args:
            session_id: Session ID
            working_memory: Working memory to preserve
            next_action: Recommended next action for next session
            error: Error that caused session to end
            tokens_consumed: Total tokens used
        """
        with session_scope() as db:
            session = (
                db.query(SessionState)
                .filter(SessionState.id == session_id)
                .first()
            )
            if not session:
                return

            session.ended_at = datetime.utcnow()
            session.working_memory = working_memory or {}
            session.next_recommended_action = next_action
            session.error_state = error
            session.tokens_consumed = tokens_consumed

            db.commit()
            self._logger.info(f"Ended session {session_id}")

    def get_last_session(self, project_id: int) -> Optional[SessionState]:
        """Get the most recent session for a project.

        Args:
            project_id: Project ID

        Returns:
            Most recent SessionState or None
        """
        with session_scope() as db:
            session = (
                db.query(SessionState)
                .filter(SessionState.project_id == project_id)
                .order_by(SessionState.started_at.desc())
                .first()
            )
            return self._detach(db, session)

    # =========================================================================
    # Progress Logging
    # =========================================================================

    def log_progress(
        self,
        project_id: int,
        action: str,
        outcome: str,
        session_id: Optional[int] = None,
        feature_id: Optional[int] = None,
        files_changed: Optional[List[str]] = None,
        git_commit: bool = False,
        tokens_used: int = 0,
    ) -> ProgressEntry:
        """Log progress on a project.

        Args:
            project_id: Project ID
            action: What was done
            outcome: What happened
            session_id: Optional session ID
            feature_id: Optional feature ID
            files_changed: List of files modified
            git_commit: Whether to create git commit
            tokens_used: Tokens consumed for this work

        Returns:
            Created ProgressEntry
        """
        with session_scope() as db:
            # Get git commit hash if committing
            commit_hash = None
            if git_commit:
                project = (
                    db.query(ProjectState)
                    .filter(ProjectState.id == project_id)
                    .first()
                )
                if project and project.git_repo_path:
                    commit_hash = self._git_commit(
                        project.git_repo_path,
                        f"{action}\n\n{outcome}",
                        files_changed or [],
                    )

            entry = ProgressEntry(
                project_id=project_id,
                session_id=session_id,
                feature_id=feature_id,
                action=action,
                outcome=outcome,
                files_changed=files_changed or [],
                git_commit_hash=commit_hash,
                tokens_used=tokens_used,
            )
            db.add(entry)
            db.commit()
            db.refresh(entry)

            self._logger.info(f"Logged progress: {action}")
            return self._detach(db, entry)

    def get_progress_log(
        self, project_id: int, limit: int = 20
    ) -> List[ProgressEntry]:
        """Get recent progress entries for a project.

        Args:
            project_id: Project ID
            limit: Maximum entries to return

        Returns:
            List of ProgressEntry instances
        """
        with session_scope() as db:
            entries = (
                db.query(ProgressEntry)
                .filter(ProgressEntry.project_id == project_id)
                .order_by(ProgressEntry.timestamp.desc())
                .limit(limit)
                .all()
            )
            return self._detach_all(db, entries)

    def get_progress_as_text(self, project_id: int, limit: int = 20) -> str:
        """Get progress log as human-readable text.

        Args:
            project_id: Project ID
            limit: Maximum entries

        Returns:
            Formatted progress log string
        """
        entries = self.get_progress_log(project_id, limit)
        if not entries:
            return "No progress recorded yet."

        lines = ["# Progress Log", ""]
        for entry in reversed(entries):  # Chronological order
            lines.append(entry.to_log_string())
            lines.append("")

        return "\n".join(lines)

    # =========================================================================
    # Decision Memory
    # =========================================================================

    def record_decision(
        self,
        project_id: int,
        context: str,
        decision: str,
        reasoning: str,
        feature_id: Optional[int] = None,
        tags: Optional[List[str]] = None,
    ) -> DecisionMemory:
        """Record a decision for future reference.

        Args:
            project_id: Project ID
            context: What situation prompted the decision
            decision: What was decided
            reasoning: Why this decision was made
            feature_id: Optional related feature
            tags: Tags for retrieval

        Returns:
            Created DecisionMemory
        """
        with session_scope() as db:
            memory = DecisionMemory(
                project_id=project_id,
                feature_id=feature_id,
                decision_context=context,
                decision_made=decision,
                reasoning=reasoning,
                tags=tags or [],
            )
            db.add(memory)
            db.commit()
            db.refresh(memory)

            self._logger.info(f"Recorded decision: {decision[:50]}...")
            return self._detach(db, memory)

    def update_decision_outcome(
        self,
        decision_id: int,
        outcome: DecisionOutcome,
        score: float,
        lesson: Optional[str] = None,
    ) -> None:
        """Update a decision with its outcome.

        Args:
            decision_id: Decision ID
            outcome: What happened
            score: -1.0 to 1.0 success score
            lesson: What was learned
        """
        with session_scope() as db:
            memory = (
                db.query(DecisionMemory)
                .filter(DecisionMemory.id == decision_id)
                .first()
            )
            if memory:
                memory.outcome = outcome
                memory.outcome_score = max(-1.0, min(1.0, score))
                memory.lesson_learned = lesson
                db.commit()

    def get_relevant_decisions(
        self,
        project_id: int,
        tags: Optional[List[str]] = None,
        limit: int = 10,
    ) -> List[DecisionMemory]:
        """Get relevant past decisions.

        Args:
            project_id: Project ID
            tags: Tags to filter by
            limit: Maximum results

        Returns:
            List of relevant decisions
        """
        with session_scope() as db:
            query = db.query(DecisionMemory).filter(
                DecisionMemory.project_id == project_id
            )

            # Note: Full-text tag search would require more complex query
            # For now, return recent decisions
            decisions = (
                query.order_by(DecisionMemory.timestamp.desc())
                .limit(limit)
                .all()
            )
            return self._detach_all(db, decisions)

    # =========================================================================
    # Git Operations
    # =========================================================================

    def _init_git_repo(self, path: Path) -> Optional[str]:
        """Initialize git repository.

        Args:
            path: Directory path

        Returns:
            Git repo path or None if failed
        """
        try:
            git_dir = path / ".git"
            if not git_dir.exists():
                subprocess.run(
                    ["git", "init"],
                    cwd=str(path),
                    check=True,
                    capture_output=True,
                )
                self._logger.info(f"Initialized git repo at {path}")
            return str(path)
        except Exception as e:
            self._logger.error(f"Failed to init git: {e}")
            return None

    def _git_commit(
        self, repo_path: str, message: str, files: List[str]
    ) -> Optional[str]:
        """Create a git commit.

        Args:
            repo_path: Repository path
            message: Commit message
            files: Files to commit (empty for all changes)

        Returns:
            Commit hash or None if failed
        """
        try:
            # Stage files
            if files:
                subprocess.run(
                    ["git", "add"] + files,
                    cwd=repo_path,
                    check=True,
                    capture_output=True,
                )
            else:
                subprocess.run(
                    ["git", "add", "-A"],
                    cwd=repo_path,
                    check=True,
                    capture_output=True,
                )

            # Commit
            subprocess.run(
                ["git", "commit", "-m", message],
                cwd=repo_path,
                check=True,
                capture_output=True,
            )

            # Get hash
            result = subprocess.run(
                ["git", "rev-parse", "HEAD"],
                cwd=repo_path,
                check=True,
                capture_output=True,
                text=True,
            )
            return result.stdout.strip()

        except Exception as e:
            self._logger.error(f"Git commit failed: {e}")
            return None

    def git_revert_to_commit(
        self, project_id: int, commit_hash: str
    ) -> bool:
        """Revert project to a specific commit.

        Args:
            project_id: Project ID
            commit_hash: Commit to revert to

        Returns:
            True if successful
        """
        project = self.get_project(project_id)
        if not project or not project.git_repo_path:
            return False

        try:
            subprocess.run(
                ["git", "reset", "--hard", commit_hash],
                cwd=project.git_repo_path,
                check=True,
                capture_output=True,
            )
            self._logger.info(f"Reverted to commit {commit_hash[:7]}")
            return True
        except Exception as e:
            self._logger.error(f"Git revert failed: {e}")
            return False

    def get_git_log(
        self, project_id: int, limit: int = 10
    ) -> List[Dict[str, str]]:
        """Get recent git commits.

        Args:
            project_id: Project ID
            limit: Maximum commits

        Returns:
            List of commit dicts with hash, message, date
        """
        project = self.get_project(project_id)
        if not project or not project.git_repo_path:
            return []

        try:
            result = subprocess.run(
                [
                    "git",
                    "log",
                    f"-{limit}",
                    "--pretty=format:%H|%s|%ai",
                ],
                cwd=project.git_repo_path,
                check=True,
                capture_output=True,
                text=True,
            )

            commits = []
            for line in result.stdout.strip().split("\n"):
                if line:
                    parts = line.split("|")
                    if len(parts) == 3:
                        commits.append(
                            {
                                "hash": parts[0],
                                "message": parts[1],
                                "date": parts[2],
                            }
                        )
            return commits

        except Exception as e:
            self._logger.error(f"Git log failed: {e}")
            return []



"""add voice support

Revision ID: ac88a4dea04b
Revises: 986e262645eb
Create Date: 2025-04-02 05:20:25.495089

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op
from sqlalchemy import inspect

from airunner.components.llm.data.chatbot import Chatbot
from airunner.components.settings.data.voice_settings import VoiceSettings
from airunner.utils.db import (
    add_table,
    add_column,
    drop_table,
    drop_column,
    create_unique_constraint,
    drop_constraint,
)


# revision identifiers, used by Alembic.
revision: str = "ac88a4dea04b"
down_revision: Union[str, None] = "986e262645eb"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    add_table(VoiceSettings)
    add_column(Chatbot, "voice_id")
    create_unique_constraint(Chatbot, ["voice_id"], "uq_chatbots_voice_id")


def downgrade() -> None:
    # Check if the constraint exists before trying to drop it
    connection = op.get_bind()
    constraint_name = "uq_chatbots_voice_id"
    inspector = inspect(connection)
    constraints = inspector.get_unique_constraints("chatbots")
    if any(
        constraint["name"] == constraint_name for constraint in constraints
    ):
        drop_constraint(Chatbot, "uq_chatbots_voice_id")
    drop_column(Chatbot, "voice_id")
    drop_table(VoiceSettings)



from PySide6.QtCore import QRectF
from PySide6.QtGui import QColor, QPen, QPainter
from PySide6.QtWidgets import QGraphicsItem
from PySide6.QtCore import QPointF

from airunner.components.application.gui.windows.main.settings_mixin import (
    SettingsMixin,
)


class GridGraphicsItem(SettingsMixin, QGraphicsItem):
    def __init__(self, view, center_point: QPointF):
        super().__init__()
        self.view = view
        self.setZValue(-100)
        self.setFlag(QGraphicsItem.ItemIgnoresTransformations, False)
        # Note: We read center_pos directly from view.center_pos in paint()
        # to ensure we always have the current value

    def boundingRect(self) -> QRectF:
        # Always cover the visible area
        rect = self.view.mapToScene(self.view.viewport().rect()).boundingRect()
        return rect

    def paint(self, painter: QPainter, option, widget=None):
        cell_size = self.view.grid_settings.cell_size
        if cell_size <= 0:
            return
        color = QColor(self.view.grid_settings.line_color)
        pen = QPen(color, self.view.grid_settings.line_width)
        painter.setPen(pen)
        visible_rect = self.boundingRect()

        # Combine canvas offset (user pan) with grid compensation (viewport adjustments)
        offset_x = (
            self.view.canvas_offset.x()
            - self.view.grid_compensation_offset.x()
        )
        offset_y = (
            self.view.canvas_offset.y()
            - self.view.grid_compensation_offset.y()
        )

        left = int(visible_rect.left())
        right = int(visible_rect.right())
        top = int(visible_rect.top())
        bottom = int(visible_rect.bottom())

        # Get the current center_pos from the view (where items are placed).
        # Grid lines should pass through center_pos and every cell_size interval.
        # 
        # To find where grid lines should be drawn in screen coordinates:
        # 1. A grid line exists at absolute positions: center_x, center_x ± cell_size, center_x ± 2*cell_size, etc.
        # 2. To convert absolute to display: display = absolute - offset
        # 3. So grid lines are at display positions: (center_x - offset) + n*cell_size
        #
        # We need to find the first grid line position >= left:
        # start_x = center_x - offset_x + n*cell_size >= left
        # n >= (left - center_x + offset_x) / cell_size
        # n = ceil((left - center_x + offset_x) / cell_size)
        # start_x = center_x - offset_x + n * cell_size
        
        try:
            center_pos = getattr(self.view, 'center_pos', None)
            center_x = int(center_pos.x()) if center_pos is not None else 0
            center_y = int(center_pos.y()) if center_pos is not None else 0
        except Exception:
            center_x = 0
            center_y = 0

        # Calculate grid line base position in display coordinates
        # Grid lines pass through (center - offset) in display space
        grid_base_x = center_x - int(offset_x)
        grid_base_y = center_y - int(offset_y)
        
        # Find the first grid line position at or before the left/top edge
        # We want: grid_base_x + n*cell_size <= left, with n being the largest such integer
        # n = floor((left - grid_base_x) / cell_size)
        import math
        n_x = math.floor((left - grid_base_x) / cell_size)
        n_y = math.floor((top - grid_base_y) / cell_size)
        
        start_x = grid_base_x + n_x * cell_size
        start_y = grid_base_y + n_y * cell_size

        # Draw vertical lines
        x = start_x
        while x <= right:
            painter.drawLine(x, top, x, bottom)
            x += cell_size
        # Draw horizontal lines
        y = start_y
        while y <= bottom:
            painter.drawLine(left, y, right, y)
            y += cell_size



"""
Video widget for video generation and playback.

Provides UI for generating videos using various models (HunyuanVideo, CogVideoX, etc.)
with real-time preview, timeline controls, and export functionality.
"""

from typing import Optional, List
import numpy as np
from PIL import Image

from PySide6.QtCore import Qt, QTimer, Slot
from PySide6.QtWidgets import (
    QFileDialog,
    QMessageBox,
    QGraphicsScene,
)
from PySide6.QtGui import QImage, QPixmap

from airunner.components.application.gui.widgets.base_widget import BaseWidget
from airunner.components.video.gui.widgets.templates.video_widget_ui import (
    Ui_video_widget,
)
from airunner.enums import SignalCode


class VideoWidget(BaseWidget):
    """
    Main video generation widget.

    Handles:
    - Video model selection
    - Prompt input and generation parameters
    - Source image selection for I2V
    - Video preview and playback
    - Timeline scrubbing
    - Progress monitoring
    - Video export
    """

    widget_class_ = Ui_video_widget

    def __init__(self, *args, **kwargs):
        """Initialize the video widget."""
        # Initialize state variables BEFORE super().__init__()
        self._current_video_frames: List[np.ndarray] = []
        self._current_frame_index: int = 0
        self._is_playing: bool = False
        self._source_image_path: Optional[str] = None
        self._generating: bool = False
        self._playback_timer = None

        # Signal handlers
        self.signal_handlers = {
            SignalCode.VIDEO_PROGRESS_SIGNAL: self.on_video_progress,
            SignalCode.VIDEO_FRAME_UPDATE_SIGNAL: self.on_frame_update,
            SignalCode.VIDEO_GENERATED_SIGNAL: self.on_video_generated,
            SignalCode.VIDEO_GENERATION_FAILED_SIGNAL: self.on_generation_failed,
        }

        # Call parent constructor
        super().__init__(*args, **kwargs)

    def initialize_ui(self) -> None:
        """Initialize UI components after widget setup."""
        try:
            super().initialize_ui()
        except Exception as e:
            self.logger.error(f"VideoWidget UI initialization failed: {e}")
            return

        try:
            self._setup_playback_timer()
            self._update_ui_state()
        except Exception as e:
            self.logger.error(
                f"VideoWidget component initialization failed: {e}"
            )

    def _setup_playback_timer(self) -> None:
        """Set up the playback timer."""
        self._playback_timer = QTimer(self)
        self._playback_timer.timeout.connect(self._advance_frame)

    def _update_ui_state(self):
        """Update UI element states based on current state."""
        if not hasattr(self, "ui") or self.ui is None:
            return

        # Enable/disable based on generation state
        is_generating = self._generating
        self.ui.btn_generate.setEnabled(not is_generating)
        self.ui.btn_cancel.setEnabled(is_generating)
        self.ui.model_selector.setEnabled(not is_generating)

        # Enable/disable based on img2vid checkbox
        is_img2vid = self.ui.checkbox_img2vid.isChecked()
        self.ui.label_source_image.setEnabled(is_img2vid)
        self.ui.lineedit_source_path.setEnabled(is_img2vid)
        self.ui.btn_browse_source.setEnabled(is_img2vid)
        self.ui.btn_clear_source.setEnabled(is_img2vid)

        # Enable/disable playback controls based on video availability
        has_video = len(self._current_video_frames) > 0
        self.ui.btn_play.setEnabled(has_video and not self._is_playing)
        self.ui.btn_pause.setEnabled(has_video and self._is_playing)
        self.ui.btn_stop.setEnabled(has_video)
        self.ui.btn_prev_frame.setEnabled(has_video)
        self.ui.btn_next_frame.setEnabled(has_video)
        self.ui.btn_export.setEnabled(has_video)
        self.ui.timeline_slider.setEnabled(has_video)

    # ===== Signal Handlers =====

    @Slot(dict)
    def on_video_progress(self, data: dict):
        """Handle video generation progress updates."""
        progress = data.get("progress", 0)
        message = data.get("message", "Generating...")

        self.ui.progress_bar.setValue(progress)
        self.ui.progress_bar.setFormat(f"%p% - {message}")

    @Slot(dict)
    def on_frame_update(self, data: dict):
        """Handle frame preview updates during generation."""
        frame = data.get("frame")
        if frame is not None:
            self._display_frame(frame)

    @Slot(dict)
    def on_video_generated(self, data: dict):
        """Handle successful video generation completion."""
        self._generating = False
        self.ui.progress_bar.setValue(100)
        self.ui.progress_bar.setFormat("Complete!")

        # Load generated video
        output_path = data.get("output_path")
        if output_path:
            self._load_video(output_path)

        self._update_ui_state()

        QMessageBox.information(
            self,
            "Video Generated",
            f"Video generated successfully!\n\n{output_path}",
        )

    @Slot(dict)
    def on_generation_failed(self, data: dict):
        """Handle video generation failure."""
        self._generating = False
        self.ui.progress_bar.setValue(0)
        self.ui.progress_bar.setFormat("Failed")
        self._update_ui_state()

        error = data.get("error", "Unknown error")
        QMessageBox.critical(
            self,
            "Generation Failed",
            f"Video generation failed:\n\n{error}",
        )

    # ===== UI Event Handlers (Auto-connected via @Slot naming convention) =====

    @Slot()
    def on_model_selector_currentIndexChanged(self):
        """Handle video model selection change (auto-connected)."""
        model_name = self.ui.model_selector.currentText()
        self.logger.info(f"Video model changed to: {model_name}")

        # Update frame limits based on model
        if "HunyuanVideo" in model_name:
            self.ui.spinbox_frames.setMaximum(513)
        elif "CogVideoX" in model_name:
            self.ui.spinbox_frames.setMaximum(129)
        elif "AnimateDiff" in model_name:
            self.ui.spinbox_frames.setMaximum(256)

    @Slot()
    def on_btn_generate_clicked(self) -> None:
        """Handle generate button click (auto-connected)."""
        # Validate inputs
        prompt = self.ui.prompt_input.toPlainText().strip()
        if not prompt:
            QMessageBox.warning(
                self,
                "Missing Prompt",
                "Please enter a prompt for video generation.",
            )
            return

        # Check for source image if img2vid is enabled
        is_img2vid = self.ui.checkbox_img2vid.isChecked()
        if is_img2vid and not self._source_image_path:
            QMessageBox.warning(
                self,
                "Missing Source Image",
                "Please select a source image for image-to-video generation.",
            )
            return

        # Gather parameters
        negative_prompt = self.ui.negative_prompt_input.toPlainText().strip()
        steps = self.ui.spinbox_steps.value()
        cfg_scale = self.ui.spinbox_cfg.value()
        num_frames = self.ui.spinbox_frames.value()
        fps = self.ui.spinbox_fps.value()
        seed = self.ui.spinbox_seed.value()
        if seed == -1:
            seed = None

        # Load source image
        init_image = None
        if is_img2vid and self._source_image_path:
            try:
                init_image = Image.open(self._source_image_path)
            except Exception as e:
                self.logger.error(f"Failed to load source image: {e}")
                QMessageBox.critical(
                    self,
                    "Image Load Error",
                    f"Failed to load source image:\n\n{e}",
                )
                return

        # Start generation
        self._generating = True
        self.ui.progress_bar.setValue(0)
        self.ui.progress_bar.setFormat("Starting generation...")
        self._update_ui_state()

        # Emit generation signal
        self.emit_signal(
            SignalCode.VIDEO_GENERATE_SIGNAL,
            {
                "prompt": prompt,
                "negative_prompt": negative_prompt,
                "init_image": init_image,
                "num_frames": num_frames,
                "fps": fps,
                "steps": steps,
                "cfg_scale": cfg_scale,
                "seed": seed,
            },
        )

    @Slot()
    def on_btn_cancel_clicked(self):
        """Handle cancel button click (auto-connected)."""
        self.emit_signal(SignalCode.INTERRUPT_VIDEO_GENERATION_SIGNAL, {})
        self._generating = False
        self.ui.progress_bar.setFormat("Cancelled")
        self._update_ui_state()

    @Slot(bool)
    def on_checkbox_img2vid_toggled(self, checked: bool):
        """Handle image-to-video checkbox toggle (auto-connected)."""
        self._update_ui_state()

    @Slot()
    def on_btn_browse_source_clicked(self):
        """Handle browse source image button click (auto-connected)."""
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Select Source Image",
            "",
            "Images (*.png *.jpg *.jpeg *.bmp *.webp);;All Files (*)",
        )

        if file_path:
            self._source_image_path = file_path
            self.ui.lineedit_source_path.setText(file_path)

    @Slot()
    def on_btn_clear_source_clicked(self):
        """Handle clear source image button click (auto-connected)."""
        self._source_image_path = None
        self.ui.lineedit_source_path.clear()

    @Slot()
    def on_btn_play_clicked(self) -> None:
        """Handle play button click (auto-connected)."""
        if not self._current_video_frames or not self._playback_timer:
            return

        self._is_playing = True
        fps = self.ui.spinbox_fps.value()
        self._playback_timer.start(int(1000 / fps))
        self._update_ui_state()

    @Slot()
    def on_btn_pause_clicked(self) -> None:
        """Handle pause button click (auto-connected)."""
        if not self._playback_timer:
            return

        self._is_playing = False
        self._playback_timer.stop()
        self._update_ui_state()

    @Slot()
    def on_btn_stop_clicked(self) -> None:
        """Handle stop button click (auto-connected)."""
        if not self._playback_timer:
            return

        self._is_playing = False
        self._playback_timer.stop()
        self._current_frame_index = 0
        self._update_frame_display()
        self._update_ui_state()

    @Slot()
    def on_btn_prev_frame_clicked(self):
        """Handle previous frame button click (auto-connected)."""
        if self._current_frame_index > 0:
            self._current_frame_index -= 1
            self._update_frame_display()

    @Slot()
    def on_btn_next_frame_clicked(self):
        """Handle next frame button click (auto-connected)."""
        if self._current_frame_index < len(self._current_video_frames) - 1:
            self._current_frame_index += 1
            self._update_frame_display()

    @Slot(int)
    def on_timeline_slider_valueChanged(self, value: int):
        """Handle timeline slider change (auto-connected)."""
        if 0 <= value < len(self._current_video_frames):
            self._current_frame_index = value
            self._update_frame_display()

    @Slot()
    def on_btn_export_clicked(self) -> None:
        """Handle export button click (auto-connected)."""
        if not self._current_video_frames:
            return

        file_path, _ = QFileDialog.getSaveFileName(
            self,
            "Export Video",
            "",
            "MP4 Video (*.mp4);;WebM Video (*.webm);;All Files (*)",
        )

        if file_path:
            # Emit export signal
            self.emit_signal(
                SignalCode.VIDEO_EXPORT_SIGNAL,
                {
                    "frames": self._current_video_frames,
                    "output_path": file_path,
                    "fps": self.ui.spinbox_fps.value(),
                },
            )

    @Slot()
    def on_btn_model_settings_clicked(self):
        """Handle model settings button click (auto-connected)."""
        # TODO: Open model settings dialog
        self.logger.info("Model settings clicked (not implemented)")

    # ===== Helper Methods =====

    def _advance_frame(self):
        """Advance to next frame during playback."""
        if not self._current_video_frames:
            self.on_btn_stop_clicked()
            return

        self._current_frame_index += 1
        if self._current_frame_index >= len(self._current_video_frames):
            # Loop playback
            self._current_frame_index = 0

        self._update_frame_display()

    def _update_frame_display(self):
        """Update the preview with current frame."""
        if not self._current_video_frames or self._current_frame_index >= len(
            self._current_video_frames
        ):
            return

        frame = self._current_video_frames[self._current_frame_index]
        self._display_frame(frame)

        # Update timeline slider
        self.ui.timeline_slider.blockSignals(True)
        self.ui.timeline_slider.setValue(self._current_frame_index)
        self.ui.timeline_slider.blockSignals(False)

        # Update timecode
        fps = self.ui.spinbox_fps.value()
        current_time = self._current_frame_index / fps
        total_time = len(self._current_video_frames) / fps
        self.ui.label_timecode.setText(
            f"{current_time:.2f}s / {total_time:.2f}s"
        )

    def _display_frame(self, frame: np.ndarray):
        """Display a frame in the preview area."""
        # Convert numpy array to QImage
        height, width, channels = frame.shape
        bytes_per_line = channels * width

        qimage = QImage(
            frame.data,
            width,
            height,
            bytes_per_line,
            QImage.Format_RGB888,
        )

        # Convert to pixmap and display
        pixmap = QPixmap.fromImage(qimage)

        # Scale to fit preview
        scene = self.ui.video_preview.scene()
        if scene is None:
            scene = QGraphicsScene()
            self.ui.video_preview.setScene(scene)

        scene.clear()
        scene.addPixmap(pixmap)
        self.ui.video_preview.fitInView(scene.sceneRect(), Qt.KeepAspectRatio)

    def _load_video(self, video_path: str):
        """Load video frames from file."""
        # TODO: Implement video loading via FFmpeg/OpenCV
        # For now, just log
        self.logger.info(f"Loading video from: {video_path}")

        # Placeholder: In a real implementation, we would:
        # 1. Use ffmpeg or opencv to extract frames
        # 2. Store frames in self._current_video_frames
        # 3. Update timeline slider range
        # 4. Display first frame

        # Update timeline
        if self._current_video_frames:
            self.ui.timeline_slider.setRange(
                0, len(self._current_video_frames) - 1
            )
            self._current_frame_index = 0
            self._update_frame_display()

        self._update_ui_state()



"""Modify default stt and tts paths

Revision ID: 7fb526dc074c
Revises: 181e31f78151
Create Date: 2025-03-07 12:14:52.042949

"""

from typing import Union
import os

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import text  # Add import for text()

from airunner.settings import AIRUNNER_BASE_PATH
from airunner.components.settings.data.path_settings import PathSettings
from airunner.utils.db import safe_alter_column

revision: str = "7fb526dc074c"
down_revision: Union[str, None] = "181e31f78151"


def upgrade() -> None:
    op.execute("DROP TABLE IF EXISTS _alembic_tmp_path_settings")

    stt_model_path = os.path.expanduser(
        os.path.join(AIRUNNER_BASE_PATH, "text", "models", "stt")
    )
    tts_model_path = os.path.expanduser(
        os.path.join(AIRUNNER_BASE_PATH, "text", "models", "tts")
    )
    safe_alter_column(
        PathSettings,
        "stt_model_path",
        existing_type=sa.String(),
        nullable=True,
        existing_server_default=stt_model_path,
    )
    safe_alter_column(
        PathSettings,
        "tts_model_path",
        existing_type=sa.String(),
        nullable=True,
        existing_server_default=tts_model_path,
    )
    update_path_values(stt_model_path, tts_model_path)


def downgrade() -> None:
    stt_model_path = os.path.expanduser(
        os.path.join(AIRUNNER_BASE_PATH, "text", "models", "stt", "models")
    )
    tts_model_path = os.path.expanduser(
        os.path.join(AIRUNNER_BASE_PATH, "text", "models", "tts", "models")
    )
    safe_alter_column(
        PathSettings,
        "stt_model_path",
        existing_type=sa.String(),
        nullable=True,
        existing_server_default=stt_model_path,
    )
    safe_alter_column(
        PathSettings,
        "tts_model_path",
        existing_type=sa.String(),
        nullable=True,
        existing_server_default=tts_model_path,
    )
    update_path_values(stt_model_path, tts_model_path)


def update_path_values(stt_model_path, tts_model_path):
    connection = op.get_bind()

    # Get all path settings
    result = connection.execute(
        text("SELECT id, base_path FROM path_settings")
    )
    path_settings = result.fetchall()

    for path_setting in path_settings:
        path_id = path_setting[0]
        base_path = path_setting[1]

        stt_model_path = stt_model_path.replace(
            os.path.expanduser(AIRUNNER_BASE_PATH), base_path
        )
        tts_model_path = tts_model_path.replace(
            os.path.expanduser(AIRUNNER_BASE_PATH), base_path
        )

        try:
            connection.execute(
                text(
                    "UPDATE path_settings SET stt_model_path = :stt, tts_model_path = :tts WHERE id = :id"
                ),
                {"stt": stt_model_path, "tts": tts_model_path, "id": path_id},
            )
            print(f"Successfully updated path_settings with ID {path_id}")
        except Exception as e:
            print(f"Error updating path_settings with ID {path_id}: {str(e)}")
            raise



"""
X4 Upscale Manager Mixins.

This package contains mixins that compose the X4UpscaleManager functionality:
- Properties: Configuration and state accessors
- Pipeline Setup: Loading and configuration
- Data Preparation: Request building and validation
- Upscaling Core: Main upscaling algorithms
- Tiling: Tile building and pasting
- Image Processing: Format conversions and fallbacks
- Response: Progress and completion signaling
- Utility: Cache and error handling
"""

from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_data_preparation_mixin import (
    X4DataPreparationMixin,
)
from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_image_processing_mixin import (
    X4ImageProcessingMixin,
)
from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_pipeline_setup_mixin import (
    X4PipelineSetupMixin,
)
from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_properties_mixin import (
    X4PropertiesMixin,
)
from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_response_mixin import (
    X4ResponseMixin,
)
from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_tiling_mixin import (
    X4TilingMixin,
)
from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_upscaling_execution_mixin import (
    X4UpscalingExecutionMixin,
)
from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_upscaling_tiling_mixin import (
    X4UpscalingTilingMixin,
)
from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_upscaling_core_mixin import (
    X4UpscalingCoreMixin,
)
from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_utility_mixin import (
    X4UtilityMixin,
)
from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_response_io_mixin import (
    X4ResponseIOMixin,
)
from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_data_utils_mixin import (
    X4DataUtilsMixin,
)
from airunner.components.art.managers.stablediffusion.x4_upscale_mixins.x4_oom_mixin import (
    X4OOMMixin,
)

__all__ = [
    "X4PropertiesMixin",
    "X4PipelineSetupMixin",
    "X4DataPreparationMixin",
    "X4DataUtilsMixin",
    "X4UpscalingCoreMixin",
    "X4UpscalingExecutionMixin",
    "X4UpscalingTilingMixin",
    "X4OOMMixin",
    "X4ResponseIOMixin",
    "X4TilingMixin",
    "X4ImageProcessingMixin",
    "X4ResponseMixin",
    "X4UtilityMixin",
]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'lora_trigger_word.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QHBoxLayout, QLabel, QPushButton,
    QSizePolicy, QWidget)

class Ui_lora_trigger_word(object):
    def setupUi(self, lora_trigger_word):
        if not lora_trigger_word.objectName():
            lora_trigger_word.setObjectName(u"lora_trigger_word")
        lora_trigger_word.resize(345, 44)
        self.horizontalLayout = QHBoxLayout(lora_trigger_word)
        self.horizontalLayout.setSpacing(5)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.trigger_word = QLabel(lora_trigger_word)
        self.trigger_word.setObjectName(u"trigger_word")

        self.horizontalLayout.addWidget(self.trigger_word)

        self.button_to_prompt = QPushButton(lora_trigger_word)
        self.button_to_prompt.setObjectName(u"button_to_prompt")
        self.button_to_prompt.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon = QIcon(QIcon.fromTheme(u"edit-paste"))
        self.button_to_prompt.setIcon(icon)

        self.horizontalLayout.addWidget(self.button_to_prompt)

        self.button_to_negative_prompt = QPushButton(lora_trigger_word)
        self.button_to_negative_prompt.setObjectName(u"button_to_negative_prompt")
        self.button_to_negative_prompt.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        self.button_to_negative_prompt.setIcon(icon)

        self.horizontalLayout.addWidget(self.button_to_negative_prompt)

        self.button_copy = QPushButton(lora_trigger_word)
        self.button_copy.setObjectName(u"button_copy")
        self.button_copy.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))
        icon1 = QIcon(QIcon.fromTheme(u"edit-copy"))
        self.button_copy.setIcon(icon1)

        self.horizontalLayout.addWidget(self.button_copy)


        self.retranslateUi(lora_trigger_word)
        self.button_to_prompt.clicked["bool"].connect(lora_trigger_word.action_click_button_to_prompt)
        self.button_to_negative_prompt.clicked.connect(lora_trigger_word.action_click_button_to_negative_prompt)
        self.button_copy.clicked.connect(lora_trigger_word.action_click_button_copy)

        QMetaObject.connectSlotsByName(lora_trigger_word)
    # setupUi

    def retranslateUi(self, lora_trigger_word):
        lora_trigger_word.setWindowTitle(QCoreApplication.translate("lora_trigger_word", u"Form", None))
        self.trigger_word.setText(QCoreApplication.translate("lora_trigger_word", u"Trigger Word", None))
#if QT_CONFIG(tooltip)
        self.button_to_prompt.setToolTip(QCoreApplication.translate("lora_trigger_word", u"Send trigger word to prompt", None))
#endif // QT_CONFIG(tooltip)
        self.button_to_prompt.setText(QCoreApplication.translate("lora_trigger_word", u"Prompt", None))
#if QT_CONFIG(tooltip)
        self.button_to_negative_prompt.setToolTip(QCoreApplication.translate("lora_trigger_word", u"Send trigger word to negative prompt", None))
#endif // QT_CONFIG(tooltip)
        self.button_to_negative_prompt.setText(QCoreApplication.translate("lora_trigger_word", u"Negative", None))
#if QT_CONFIG(tooltip)
        self.button_copy.setToolTip(QCoreApplication.translate("lora_trigger_word", u"Copy trigger word to clipboard", None))
#endif // QT_CONFIG(tooltip)
        self.button_copy.setText("")
    # retranslateUi




"""Unit tests for GenerationMixin."""

from unittest.mock import Mock, patch
import pytest
from langchain_core.messages import AIMessage

from airunner.components.llm.managers.mixins.generation_mixin import (
    GenerationMixin,
)
from airunner.components.llm.managers.llm_request import LLMRequest
from airunner.components.llm.managers.llm_response import LLMResponse
from airunner.enums import LLMActionType


class TestableGenerationMixin(GenerationMixin):
    """Testable version of GenerationMixin with required dependencies."""

    def __init__(self):
        """Initialize with mock dependencies."""
        self.logger = Mock()
        self.api = Mock()
        self._workflow_manager = Mock()
        self._tool_manager = Mock()
        self._interrupted = False
        self._current_model_path = "/path/to/model"
        self.model_path = "/path/to/model"
        self.llm_generator_settings = Mock()
        self.chatbot = Mock()
        self.load = Mock()
        self.unload = Mock()

    def get_system_prompt_for_action(self, action: LLMActionType) -> str:
        """Mock method for getting system prompt."""
        return "Default system prompt"

    def get_system_prompt_with_context(
        self, action: LLMActionType, tool_categories=None, force_tool=None
    ) -> str:
        """Mock method for getting context-aware system prompt."""
        return "Default system prompt"


@pytest.fixture
def mixin():
    """Create a testable GenerationMixin instance."""
    return TestableGenerationMixin()


@pytest.fixture
def llm_request():
    """Create a test LLM request."""
    return LLMRequest(node_id="test-node")


class TestSetupGenerationWorkflow:
    """Tests for _setup_generation_workflow method."""

    def test_uses_provided_system_prompt(self, mixin):
        """Should use provided system prompt when given."""
        custom_prompt = "Custom system prompt"
        result = mixin._setup_generation_workflow(
            LLMActionType.CHAT, custom_prompt
        )

        assert result == custom_prompt
        mixin._workflow_manager.update_system_prompt.assert_called_once_with(
            custom_prompt
        )

    def test_uses_action_system_prompt_when_none_provided(self, mixin):
        """Should get action prompt when none provided."""
        result = mixin._setup_generation_workflow(LLMActionType.CHAT, None)

        assert result == "Default system prompt"
        mixin._workflow_manager.update_system_prompt.assert_called_once_with(
            "Default system prompt"
        )

    def test_updates_tools_from_tool_manager(self, mixin):
        """Should update workflow tools from tool manager."""
        mock_tools = [Mock(), Mock()]
        mixin._tool_manager.get_tools_for_action.return_value = mock_tools

        mixin._setup_generation_workflow(LLMActionType.CODE, None)

        mixin._tool_manager.get_tools_for_action.assert_called_once_with(
            LLMActionType.CODE
        )
        mixin._workflow_manager.update_tools.assert_called_once_with(
            mock_tools
        )

    def test_handles_missing_workflow_manager(self, mixin):
        """Should handle missing workflow manager gracefully."""
        mixin._workflow_manager = None

        result = mixin._setup_generation_workflow(LLMActionType.CHAT, None)

        assert result == "Default system prompt"

    def test_handles_missing_tool_manager(self, mixin):
        """Should handle missing tool manager gracefully."""
        mixin._tool_manager = None

        result = mixin._setup_generation_workflow(
            LLMActionType.CHAT, "Test prompt"
        )

        assert result == "Test prompt"
        mixin._workflow_manager.update_system_prompt.assert_called_once_with(
            "Test prompt"
        )


class TestCreateStreamingCallback:
    """Tests for _create_streaming_callback method."""

    def test_callback_accumulates_response(self, mixin, llm_request):
        """Should accumulate response text."""
        complete_response = [""]
        sequence_counter = [0]

        callback = mixin._create_streaming_callback(
            llm_request, complete_response, sequence_counter
        )

        callback("Hello")
        callback(" world")

        assert complete_response[0] == "Hello world"

    def test_callback_increments_sequence(self, mixin, llm_request):
        """Should increment sequence counter for each token."""
        complete_response = [""]
        sequence_counter = [0]

        callback = mixin._create_streaming_callback(
            llm_request, complete_response, sequence_counter
        )

        callback("Hello")
        callback(" world")
        callback("!")

        assert sequence_counter[0] == 3

    def test_callback_sends_llm_signals(self, mixin, llm_request):
        """Should send LLM signals for each token."""
        complete_response = [""]
        sequence_counter = [0]

        callback = mixin._create_streaming_callback(
            llm_request, complete_response, sequence_counter
        )

        callback("Hello")

        mixin.api.llm.send_llm_text_streamed_signal.assert_called_once()
        call_args = mixin.api.llm.send_llm_text_streamed_signal.call_args[0][0]
        assert isinstance(call_args, LLMResponse)
        assert call_args.message == "Hello"
        assert call_args.is_end_of_message is False
        assert call_args.is_first_message is True
        assert call_args.sequence_number == 1

    def test_callback_marks_first_message(self, mixin, llm_request):
        """Should mark first message correctly."""
        complete_response = [""]
        sequence_counter = [0]

        callback = mixin._create_streaming_callback(
            llm_request, complete_response, sequence_counter
        )

        callback("First")
        first_call = (
            mixin.api.llm.send_llm_text_streamed_signal.call_args_list[0][0][0]
        )

        callback("Second")
        second_call = (
            mixin.api.llm.send_llm_text_streamed_signal.call_args_list[1][0][0]
        )

        assert first_call.is_first_message is True
        assert second_call.is_first_message is False

    def test_callback_ignores_empty_tokens(self, mixin, llm_request):
        """Should ignore empty token strings."""
        complete_response = [""]
        sequence_counter = [0]

        callback = mixin._create_streaming_callback(
            llm_request, complete_response, sequence_counter
        )

        callback("")

        assert complete_response[0] == ""
        assert sequence_counter[0] == 0
        mixin.api.llm.send_llm_text_streamed_signal.assert_not_called()

    def test_callback_includes_node_id(self, mixin, llm_request):
        """Should include node_id in signals."""
        complete_response = [""]
        sequence_counter = [0]

        callback = mixin._create_streaming_callback(
            llm_request, complete_response, sequence_counter
        )

        callback("Test")

        call_args = mixin.api.llm.send_llm_text_streamed_signal.call_args[0][0]
        assert call_args.node_id == "test-node"

    def test_callback_handles_none_request(self, mixin):
        """Should handle None request gracefully."""
        complete_response = [""]
        sequence_counter = [0]

        callback = mixin._create_streaming_callback(
            None, complete_response, sequence_counter
        )

        callback("Test")

        call_args = mixin.api.llm.send_llm_text_streamed_signal.call_args[0][0]
        assert call_args.node_id is None


class TestHandleInterruptedGeneration:
    """Tests for _handle_interrupted_generation method."""

    def test_logs_interruption(self, mixin, llm_request):
        """Should log interruption message."""
        mixin._handle_interrupted_generation(llm_request, 5)

        mixin.logger.info.assert_called_once_with(
            "Generation interrupted by user"
        )

    def test_returns_empty_string(self, mixin, llm_request):
        """Should return empty string (no visible interrupt message)."""
        result = mixin._handle_interrupted_generation(llm_request, 5)

        assert result == ""

    def test_sends_interrupt_signal(self, mixin, llm_request):
        """Should send interrupt signal with correct parameters."""
        mixin._handle_interrupted_generation(llm_request, 5)

        mixin.api.llm.send_llm_text_streamed_signal.assert_called_once()
        call_args = mixin.api.llm.send_llm_text_streamed_signal.call_args[0][0]
        assert isinstance(call_args, LLMResponse)
        assert call_args.message == ""
        assert call_args.is_end_of_message is True
        assert call_args.sequence_number == 6  # counter + 1

    def test_handles_none_request(self, mixin):
        """Should handle None request gracefully."""
        result = mixin._handle_interrupted_generation(None, 3)

        assert result == ""
        call_args = mixin.api.llm.send_llm_text_streamed_signal.call_args[0][0]
        assert call_args.node_id is None


class TestHandleGenerationError:
    """Tests for _handle_generation_error method."""

    def test_logs_error(self, mixin, llm_request):
        """Should log error with exc_info."""
        exc = ValueError("Test error")

        mixin._handle_generation_error(exc, llm_request)

        mixin.logger.error.assert_called_once()
        call_args = mixin.logger.error.call_args
        assert "Error during generation: Test error" in str(call_args)
        assert call_args[1]["exc_info"] is True

    def test_returns_error_message(self, mixin, llm_request):
        """Should return formatted error message."""
        exc = ValueError("Test error")

        result = mixin._handle_generation_error(exc, llm_request)

        assert result == "Error: Test error"

    def test_sends_error_signal(self, mixin, llm_request):
        """Should send error signal."""
        exc = ValueError("Test error")

        mixin._handle_generation_error(exc, llm_request)

        mixin.api.llm.send_llm_text_streamed_signal.assert_called_once()
        call_args = mixin.api.llm.send_llm_text_streamed_signal.call_args[0][0]
        assert isinstance(call_args, LLMResponse)
        assert call_args.message == "Error: Test error"
        assert call_args.is_end_of_message is False

    def test_handles_none_request(self, mixin):
        """Should handle None request gracefully."""
        exc = ValueError("Test error")

        result = mixin._handle_generation_error(exc, None)

        assert result == "Error: Test error"
        call_args = mixin.api.llm.send_llm_text_streamed_signal.call_args[0][0]
        assert call_args.node_id is None


class TestExtractFinalResponse:
    """Tests for _extract_final_response method."""

    def test_extracts_content_from_ai_message(self, mixin):
        """Should extract content from last AIMessage."""
        ai_msg = AIMessage(content="Final response")
        result = {"messages": [ai_msg]}

        extracted = mixin._extract_final_response(result)

        assert extracted == "Final response"

    def test_returns_empty_for_no_messages(self, mixin):
        """Should return empty string when no messages."""
        result = {"messages": []}

        extracted = mixin._extract_final_response(result)

        assert extracted == ""

    def test_returns_empty_for_missing_messages_key(self, mixin):
        """Should return empty string when messages key missing."""
        result = {}

        extracted = mixin._extract_final_response(result)

        assert extracted == ""

    def test_returns_empty_for_none_result(self, mixin):
        """Should return empty string for None result."""
        extracted = mixin._extract_final_response(None)

        assert extracted == ""

    def test_uses_last_ai_message(self, mixin):
        """Should use last AIMessage when multiple exist."""
        msg1 = AIMessage(content="First")
        msg2 = AIMessage(content="Second")
        msg3 = AIMessage(content="Third")
        result = {"messages": [msg1, msg2, msg3]}

        extracted = mixin._extract_final_response(result)

        assert extracted == "Third"

    def test_filters_non_ai_messages(self, mixin):
        """Should filter out non-AIMessage types."""
        from langchain_core.messages import HumanMessage

        human = HumanMessage(content="Human message")
        ai = AIMessage(content="AI message")
        result = {"messages": [human, ai, human]}

        extracted = mixin._extract_final_response(result)

        assert extracted == "AI message"

    def test_handles_empty_content(self, mixin):
        """Should return empty for AIMessage with no content."""
        ai_msg = AIMessage(content="")
        result = {"messages": [ai_msg]}

        extracted = mixin._extract_final_response(result)

        assert extracted == ""

    def test_handles_none_content(self, mixin):
        """Should return empty for AIMessage with None content."""
        ai_msg = AIMessage(content="")
        ai_msg.content = None  # Set to None after creation
        result = {"messages": [ai_msg]}

        extracted = mixin._extract_final_response(result)

        assert extracted == ""


class TestDoGenerate:
    """Tests for _do_generate method."""

    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_basic_generation_flow(self, mock_torch, mixin, llm_request):
        """Should execute basic generation successfully."""
        mock_torch.cuda.is_available.return_value = True
        mixin._workflow_manager.stream.return_value = [
            AIMessage(content="Test response")
        ]

        result = mixin._do_generate(
            "Test prompt", LLMActionType.CHAT, llm_request=llm_request
        )

        assert result["response"] == "Test response"

    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_reloads_model_on_path_mismatch(self, mock_torch, mixin):
        """Should reload model when path mismatch detected."""
        mixin._current_model_path = "/old/path"
        mixin.model_path = "/new/path"
        mixin._workflow_manager.stream.return_value = []

        mixin._do_generate("Test", LLMActionType.CHAT)

        mixin.unload.assert_called_once()
        mixin.load.assert_called_once()

    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_creates_default_request_if_none(self, mock_torch, mixin):
        """Should create default LLMRequest if None provided."""
        mixin._workflow_manager.stream.return_value = []

        mixin._do_generate("Test", LLMActionType.CHAT, llm_request=None)

        # Should not raise exception
        assert True

    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_sets_token_callback(self, mock_torch, mixin):
        """Should set and clear token callback."""
        mixin._workflow_manager.stream.return_value = []

        mixin._do_generate("Test", LLMActionType.CHAT)

        # Should be called twice: once to set, once to clear
        assert mixin._workflow_manager.set_token_callback.call_count == 2

    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_resets_interrupted_flag(self, mock_torch, mixin):
        """Should reset interrupted flag before generation."""
        mixin._interrupted = True
        mixin._workflow_manager.stream.return_value = []

        mixin._do_generate("Test", LLMActionType.CHAT)

        assert mixin._interrupted is False

    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_handles_interrupt_during_streaming(self, mock_torch, mixin):
        """Should handle interrupt during streaming."""

        def stream_with_interrupt(*args, **kwargs):
            mixin._interrupted = True
            yield AIMessage(content="Partial")

        mixin._workflow_manager.stream = stream_with_interrupt

        result = mixin._do_generate("Test", LLMActionType.CHAT)

        # Interrupt should end cleanly without adding visible text
        assert result["response"] == ""

    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_handles_generation_error(self, mock_torch, mixin):
        """Should handle generation errors gracefully."""
        mixin._workflow_manager.stream.side_effect = ValueError("Test error")

        result = mixin._do_generate("Test", LLMActionType.CHAT)

        assert result["response"] == "Error: Test error"

    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_clears_cuda_cache(self, mock_torch, mixin):
        """Should clear CUDA cache when available."""
        mock_torch.cuda.is_available.return_value = True
        mixin._workflow_manager.stream.return_value = []

        mixin._do_generate("Test", LLMActionType.CHAT)

        mock_torch.cuda.empty_cache.assert_called_once()
        mock_torch.cuda.synchronize.assert_called_once()

    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_sends_end_of_message_signal(self, mock_torch, mixin, llm_request):
        """Should send end of message signal."""
        mixin._workflow_manager.stream.return_value = []

        mixin._do_generate("Test", LLMActionType.CHAT, llm_request=llm_request)

        # Should send at least one end-of-message signal
        calls = mixin.api.llm.send_llm_text_streamed_signal.call_args_list
        end_calls = [c for c in calls if c[0][0].is_end_of_message is True]
        assert len(end_calls) > 0

    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_returns_error_when_no_workflow_manager(self, mock_torch, mixin):
        """Should return error when workflow manager missing."""
        mixin._workflow_manager = None

        result = mixin._do_generate("Test", LLMActionType.CHAT)

        assert result["response"] == "Error: workflow unavailable"


class TestSendFinalMessage:
    """Tests for _send_final_message method."""

    def test_sends_end_of_message_signal(self, mixin, llm_request):
        """Should send end-of-message signal."""
        mixin._send_final_message(llm_request)

        mixin.api.llm.send_llm_text_streamed_signal.assert_called_once()
        call_args = mixin.api.llm.send_llm_text_streamed_signal.call_args[0][0]
        assert isinstance(call_args, LLMResponse)
        assert call_args.is_end_of_message is True

    def test_includes_node_id(self, mixin, llm_request):
        """Should include node_id in signal."""
        mixin._send_final_message(llm_request)

        call_args = mixin.api.llm.send_llm_text_streamed_signal.call_args[0][0]
        assert call_args.node_id == "test-node"

    def test_handles_none_request(self, mixin):
        """Should handle None request gracefully."""
        mixin._send_final_message(None)

        call_args = mixin.api.llm.send_llm_text_streamed_signal.call_args[0][0]
        assert call_args.node_id is None


class TestDoSetSeed:
    """Tests for _do_set_seed method."""

    @patch("airunner.components.llm.managers.mixins.generation_mixin.random")
    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_uses_override_settings(self, mock_torch, mock_random, mixin):
        """Should use override settings when enabled."""
        mixin.llm_generator_settings.override_parameters = True
        mixin.llm_generator_settings.seed = 42
        mixin.llm_generator_settings.random_seed = False
        mock_torch.cuda.is_available.return_value = True

        mixin._do_set_seed()

        mock_torch.manual_seed.assert_called_once_with(42)
        mock_random.seed.assert_called_once_with(42)

    @patch("airunner.components.llm.managers.mixins.generation_mixin.random")
    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_uses_chatbot_settings(self, mock_torch, mock_random, mixin):
        """Should use chatbot settings when override disabled."""
        mixin.llm_generator_settings.override_parameters = False
        mixin.chatbot.seed = 99
        mixin.chatbot.random_seed = False
        mock_torch.cuda.is_available.return_value = True

        mixin._do_set_seed()

        mock_torch.manual_seed.assert_called_once_with(99)
        mock_random.seed.assert_called_once_with(99)

    @patch("airunner.components.llm.managers.mixins.generation_mixin.random")
    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_sets_cuda_seeds(self, mock_torch, mock_random, mixin):
        """Should set CUDA seeds when CUDA available."""
        mixin.llm_generator_settings.override_parameters = True
        mixin.llm_generator_settings.seed = 42
        mixin.llm_generator_settings.random_seed = False
        mock_torch.cuda.is_available.return_value = True

        mixin._do_set_seed()

        mock_torch.cuda.manual_seed_all.assert_called_once_with(42)

    @patch("airunner.components.llm.managers.mixins.generation_mixin.random")
    @patch("airunner.components.llm.managers.mixins.generation_mixin.torch")
    def test_skips_seed_when_random_enabled(
        self, mock_torch, mock_random, mixin
    ):
        """Should not set seed when random_seed enabled."""
        mixin.llm_generator_settings.override_parameters = True
        mixin.llm_generator_settings.seed = 42
        mixin.llm_generator_settings.random_seed = True
        mock_torch.cuda.is_available.return_value = False

        mixin._do_set_seed()

        # Should not be called with the seed value
        mock_torch.manual_seed.assert_not_called()
        mock_random.seed.assert_not_called()
